# https://github.com/NicolasZucchet/minimal-LRU/blob/main/lru/model.py
from functools import partial
from beartype.typing import Optional, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Complex, Float, PRNGKeyArray, Scalar, Shaped, jaxtyped

from memax.mtypes import Input, StartFlag
from memax.linen.groups import Semigroup, Resettable
from memax.linen.gras import GRAS
from memax.linen.scans import semigroup_scan

LRURecurrentState = Tuple[Complex[Array, "Recurrent"], Complex[Array, "Recurrent"]]
LRURecurrentStateWithReset = Tuple[LRURecurrentState, StartFlag]


# Inits
@jaxtyped(typechecker=typechecker)
def glorot_init(
    key: PRNGKeyArray, shape: Tuple[int, ...], normalization: Scalar = jnp.array(1.0)
):
    return jax.random.normal(key=key, shape=shape) / normalization


@jaxtyped(typechecker=typechecker)
def nu_init(
    key: PRNGKeyArray,
    shape: Tuple[int, ...],
    r_min: float,
    r_max: float,
    dtype: jnp.dtype = jnp.float32,
) -> jax.Array:
    u = jax.random.uniform(key=key, shape=shape, dtype=dtype)
    return jnp.log(-0.5 * jnp.log(u * (r_max**2 - r_min**2) + r_min**2))


@jaxtyped(typechecker=typechecker)
def theta_init(
    key: PRNGKeyArray,
    shape: Tuple[int, ...],
    max_phase: float,
    dtype: jnp.dtype = jnp.float32,
) -> jax.Array:
    u = jax.random.uniform(key, shape=shape, dtype=dtype)
    return jnp.log(max_phase * u)


@jaxtyped(typechecker=typechecker)
def diag_lambda(nu_log: jax.Array, theta_log: jax.Array) -> jax.Array:
    return jnp.exp(-jnp.exp(nu_log) + 1j * jnp.exp(theta_log))


@jaxtyped(typechecker=typechecker)
def gamma_log_init(
    key: PRNGKeyArray, nu_log: jax.Array, theta_log: jax.Array
) -> jax.Array:
    return jnp.log(jnp.sqrt(1 - jnp.abs(diag_lambda(nu_log, theta_log)) ** 2))


class LRUSemigroup(Semigroup):
    """The Linear Recurrent Unit semigroup (recurrent update) from https://arxiv.org/abs/2303.06349."""

    recurrent_size: int

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LRURecurrentState:
        # Represent a diagonal matrix as a vector
        return (
            jnp.ones((self.recurrent_size,), dtype=jnp.complex64),
            jnp.zeros((self.recurrent_size), dtype=jnp.complex64),
        )

    @nn.nowrap
    def zero_carry(self) -> LRURecurrentState:
        return (
            jnp.zeros((self.recurrent_size,), dtype=jnp.complex64),
            jnp.zeros((self.recurrent_size), dtype=jnp.complex64),
        )

    @jaxtyped(typechecker=typechecker)
    @nn.compact
    def __call__(
        self, carry: LRURecurrentState, input: LRURecurrentState
    ) -> LRURecurrentState:
        # Ax + Bu, but A is diagonal, and we treat it as a vector
        # So we can be more efficient by writing Ax as vec(A) * x
        A_i, bu_i = carry
        A_j, bu_j = input
        return A_j * A_i, A_j * bu_i + bu_j
        # return carry * self.diag_lambda() + input


class LRU(GRAS):
    """
    The Linear Recurrent Unit from https://arxiv.org/abs/2303.06349.

    You might want to use this as a building block for a more complex model.
    """

    hidden_size: int  # output dimensions
    recurrent_size: int  # hidden state dimension
    r_min: float = 0.0
    r_max: float = 1.0
    max_phase: float = jnp.pi * 2

    def setup(self):
        self.B_re = self.param(
            "B_re",
            partial(glorot_init, normalization=jnp.sqrt(2 * self.hidden_size)),
            (self.recurrent_size, self.hidden_size),
        )
        self.B_im = self.param(
            "B_im",
            partial(glorot_init, normalization=jnp.sqrt(2 * self.hidden_size)),
            (self.recurrent_size, self.hidden_size),
        )
        self.C_re = self.param(
            "C_re",
            partial(glorot_init, normalization=jnp.sqrt(self.recurrent_size)),
            (self.hidden_size, self.recurrent_size),
        )
        self.C_im = self.param(
            "C_im",
            partial(glorot_init, normalization=jnp.sqrt(self.recurrent_size)),
            (self.hidden_size, self.recurrent_size),
        )
        self.D = self.param("D", glorot_init, (self.hidden_size,))
        self.nu_log = self.param(
            "nu_log",
            partial(nu_init, r_min=self.r_min, r_max=self.r_max),
            (self.recurrent_size,),
        )
        self.theta_log = self.param(
            "theta_log",
            partial(theta_init, max_phase=self.max_phase),
            (self.recurrent_size,),
        )
        self.gamma_log = self.param(
            "gamma_log",
            partial(gamma_log_init, nu_log=self.nu_log, theta_log=self.theta_log),
        )

    @jaxtyped(typechecker=typechecker)
    def diag_lambda(self) -> Complex[Array, "Recurrent"]:
        return jnp.exp(-jnp.exp(self.nu_log) + 1j * jnp.exp(self.theta_log))

    @jaxtyped(typechecker=typechecker)
    def forward_map(self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None):
        emb, start = x
        B_norm = jax.lax.complex(self.B_re, self.B_im) * jnp.expand_dims(
            jnp.exp(self.gamma_log), axis=-1
        )
        Bu = B_norm @ emb.astype(jnp.complex64)
        return (diag_lambda(self.nu_log, self.theta_log), Bu), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: LRURecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        state, reset_flag = h
        emb, start = x
        lambdas, lambda_x_Bu = state
        out = (jax.lax.complex(self.C_re, self.C_im) @ lambda_x_Bu).real + self.D * emb
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LRURecurrentStateWithReset:
        return self.algebra.initialize_carry(key)

    @nn.nowrap
    def zero_carry(self) -> LRURecurrentStateWithReset:
        return self.algebra.zero_carry()

    @staticmethod
    def default_algebra(**kwargs):
        return Resettable(LRUSemigroup(**kwargs))

    @staticmethod
    def default_scan():
        return semigroup_scan



from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

MLPRecurrentState = Float[Array, "0"] # Empty array because MLP is not recurrent
MLPRecurrentStateWithReset = Tuple[MLPRecurrentState, StartFlag]


class MLPSemigroup(Semigroup):

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> MLPRecurrentState:
        return jnp.zeros((0,))

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: MLPRecurrentState, input: MLPRecurrentState
    ) -> MLPRecurrentState:
        return jnp.zeros((0,))


class MLP(GRAS):
    """A simple non-recurrent MLP. This is useful for debug purposes,
    or comparing a sequence model against a memory-free baseline.

    You might want to use this as a building block for a more complex model.
    """

    recurrent_size: int
    scan: Callable[
        [
            Callable[
                [MLPRecurrentStateWithReset, MLPRecurrentStateWithReset],
                MLPRecurrentStateWithReset,
            ],
            MLPRecurrentStateWithReset,
            MLPRecurrentStateWithReset,
        ],
        MLPRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra

    project: nn.Linear

    def __init__(self, recurrent_size, key):
        self.recurrent_size = recurrent_size
        self.algebra = Resettable(MLPSemigroup())
        self.scan = semigroup_scan

        keys = jax.random.split(key)

        self.project = nn.Sequential(
            [
                nn.Linear(recurrent_size, recurrent_size, key=keys[1]),
                nn.Lambda(jax.nn.leaky_relu),
            ]
        )

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> MLPRecurrentStateWithReset:
        _, start = x
        return (jnp.zeros((0,)), start)

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: MLPRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        emb, start = x
        none, reset_carry = h
        return self.project(emb)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> MLPRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped, Bool

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan
from memax.utils import combine_and_right_align

StackRecurrentState = Tuple[Float[Array, "Stack Recurrent"], Bool[Array, "Stack"]]
StackRecurrentStateWithReset = Tuple[StackRecurrentState, StartFlag]


class StackSemigroup(Semigroup):
    """A sliding window semigroup example.

    This allows you to define recurrent function that, for example,
    rely on the two (or more) most recent inputs through a sliding window.

    Applications include dot-product attention, RWKV, etc.
    """
    stack_size: int
    recurrent_size: int

    def __init__(self, recurrent_size: int, stack_size: int):
        self.stack_size = stack_size
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> StackRecurrentState:
        stack = jnp.zeros((self.stack_size, self.recurrent_size))
        #return stack
        # Valid (non-pad) mask
        mask = jnp.zeros((self.stack_size,), dtype=bool)
        return (stack, mask)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: StackRecurrentState, input: StackRecurrentState
    ) -> StackRecurrentState:
        # We would like to do the below
        # But cannot due to concretization error
        # Caused by dynamic indexing (mask)
        #
        # left = cstack[cmask]
        # right = stack[mask]

        # mleft = cmask[cmask]
        # mright = mask[mask]

        # stack = jnp.concatenate([left, right])[-stack_size:]
        # mask = jnp.concatenate([mleft, mright])[-stack_size:]

        # So we use a tricky function instead
        cstack, cmask = carry
        stack, mask = input
        out_stack, out_mask = combine_and_right_align(cstack, cmask, stack, mask)
        return (out_stack, out_mask)


class Stack(GRAS):
    """A fairly straightforward "recurrent" model that merely keeps a sliding
    window of the past. You may use this model as a blueprint to implement 
    "less recurrent" models that rely on more than the current input and recurrent state.

    Applications include dot-product attention, RWKV, etc.
    """

    recurrent_size: int
    stack_size: int
    scan: Callable[
        [
            Callable[
                [StackRecurrentStateWithReset, StackRecurrentStateWithReset],
                StackRecurrentStateWithReset,
            ],
            StackRecurrentStateWithReset,
            StackRecurrentStateWithReset,
        ],
        StackRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra

    g: nn.Sequential

    def __init__(self, recurrent_size, stack_size, key):
        self.recurrent_size = recurrent_size
        self.stack_size = stack_size
        self.algebra = Resettable(StackSemigroup(recurrent_size, stack_size=stack_size))
        self.scan = semigroup_scan
        self.g = nn.Linear(recurrent_size * stack_size, recurrent_size, key=key)

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> StackRecurrentStateWithReset:
        emb, start = x
        # Add stack dim for concat
        #return emb.reshape(1, -1), start
        mask = jnp.concatenate([
            jnp.zeros((self.stack_size - 1), dtype=bool),
            jnp.ones((1,), dtype=bool)
        ])
        emb = jnp.concatenate([
            jnp.zeros((self.stack_size - 1, *emb.shape), dtype=emb.dtype),
            emb.reshape(1, -1)
        ])
        return (emb, mask), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: StackRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        emb, start = x
        state, reset_carry = h
        z, mask = state
        # You can do something more intelligent with masking if needed
        z = self.g(z.reshape(-1))
        return z

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> StackRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



def test_readme():
    import equinox as eqx
    import jax
    import jax.numpy as jnp

    from memax.equinox.set_actions.gru import GRU
    from memax.equinox.models.residual import ResidualModel
    from memax.equinox.semigroups.lru import LRU, LRUSemigroup
    from memax.utils import debug_shape

    # You can pack multiple subsequences into a single sequence using the start flag
    sequence_starts = jnp.array([True, False, False, True, False])
    x = jnp.zeros((5, 3))
    inputs = (x, sequence_starts)

    # Initialize a multi-layer recurrent model
    key = jax.random.key(0)
    make_layer_fn = lambda recurrent_size, key: LRU(
        hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key
    )
    model = ResidualModel(
        make_layer_fn=make_layer_fn,
        input_size=3,
        recurrent_size=16,
        output_size=4,
        num_layers=2,
        key=key,
    )

    # Note: We also have layers if you want to build your own model
    layer = LRU(hidden_size=16, recurrent_size=16, key=key)
    # Or semigroups/set actions (scanned functions) if you want to build your own layer
    sg = LRUSemigroup(recurrent_size=16)

    # Run the model! All models are jit-capable, using equinox.filter_jit
    h = eqx.filter_jit(model.initialize_carry)()
    # Unlike most other libraries, we output ALL recurrent states h, not just the most recent
    h, y = eqx.filter_jit(model)(h, inputs)
    # Since we have two layers, we have a recurrent state of shape
    print(debug_shape(h))
    #     ((5, 16), # Recurrent states of first layer
    #     (5,) # Start carries for first layer
    #     (5, 16) # Recurrent states of second layer
    #     (5,)) # Start carries for second layer
    # 
    # Do your prediction
    prediction = jax.nn.softmax(y)

    # If you want to continue rolling out the RNN from h[-1]
    # you should use the following helper function to extract
    # h[-1] from the nested recurrent state
    latest_h = eqx.filter_jit(model.latest_recurrent_state)(h)
    # Continue rolling out as you please! You can use a single timestep
    # or another sequence.
    last_h, last_y = eqx.filter_jit(model)(latest_h, inputs)

    # We can use a similar approach with RNNs
    make_layer_fn = lambda recurrent_size, key: GRU(
        recurrent_size=recurrent_size, key=key
    )
    model = ResidualModel(
        make_layer_fn=make_layer_fn,
        input_size=3,
        recurrent_size=16,
        output_size=4,
        num_layers=2,
        key=jax.random.key(0),
    )
    h = eqx.filter_jit(model.initialize_carry)()
    h, y = eqx.filter_jit(model)(h, inputs)
    prediction = jax.nn.softmax(y)
    latest_h = eqx.filter_jit(model.latest_recurrent_state)(h)
    h, y = eqx.filter_jit(model)(latest_h, inputs)

def test_readme_quickstart():
    from memax.equinox.train_utils import get_residual_memory_models
    import jax
    import jax.numpy as jnp
    from equinox import filter_jit, filter_vmap
    from memax.equinox.train_utils import add_batch_dim

    T, F = 5, 6 # time and feature dim

    model = get_residual_memory_models(
        input=F, hidden=8, output=1, num_layers=2, 
        models=["LRU"], key=jax.random.key(0)
    )["LRU"]

    starts = jnp.array([True, False, False, True, False])
    xs = jnp.zeros((T, F)) 
    hs, ys = filter_jit(model)(model.initialize_carry(), (xs, starts))
    last_h = filter_jit(model.latest_recurrent_state)(hs)

    # with batch dim
    B = 4
    starts = jnp.zeros((B, T), dtype=bool)
    xs = jnp.zeros((B, T, F))
    hs_0 = add_batch_dim(model.initialize_carry(), B)
    hs, ys = filter_jit(filter_vmap(model))(hs_0, (xs, starts))

if __name__ == '__main__':
    test_readme()
    test_readme_quickstart()


"""Tests the framestacking method works with associative scan"""
import equinox as eqx
import jax
import jax.numpy as jnp
import pytest

from memax.equinox.semigroups.stack import Stack


def test_stack():
    T = 4
    F = 2
    stack_size = 3

    m = Stack(recurrent_size=F, stack_size=stack_size, key=jax.random.key(0))
    x = jnp.arange(T * F, dtype=jnp.float32).reshape((T, F))
    starts = jnp.zeros((T,), dtype=bool)
    inputs = (x, starts)

    h = m.initialize_carry()
    hs, ys = m(h, inputs)
    masks = hs[0][1]
    states = hs[0][0]
    # mask == [[0,0,1], [0, 1, 1], [1, 1, 1] ... ]
    assert jnp.allclose(
        masks,
        jnp.array([
            [False, False, True],
            [False, True, True],
            [True, True, True],
            [True, True, True],
        ])
    )
    # states = [ [0, 0, x[0]], [0, x[0], x[1]], ...]
    zero = jnp.zeros_like(x[0])
    tgt = jnp.array([
        [zero, zero, x[0]],
        [zero, x[0], x[1]],
        [x[0], x[1], x[2]],
        [x[1], x[2], x[3]],
    ])
    assert jnp.allclose(
        states,
        tgt
    ), f"{tgt}!=\n{states}"

if __name__ == "__main__":
    test_stack()


from beartype.typing import Callable, Optional, Tuple

import flax.linen as nn
import jax
from jaxtyping import PRNGKeyArray, Shaped

from memax.mtypes import Input, OutputEmbedding, RecurrentState, SingleRecurrentState
from memax.linen.groups import BinaryAlgebra, Module


class GRAS(Module):
    r"""A Generalized Recurrent Algebraic Structure (GRAS). Given a recurrent state and inputs, returns the corresponding recurrent states and outputs

    # Generalized Recurrent Algebraic Structure (GRAS)

    A GRAS contains a **set action** $(H, Z, \bullet)$, an initial state $h_0 \in H$, and two maps/functions 
    
    $f$ maps input features and a boolean start flag to the action space $Z$. 

    $f: X^n \times \\{0, 1\\}^n \mapsto Z^n$

    $\bullet$ applies an action $z \in Z$ to the recurrent state $h \in H$.

    $\bullet: H \times Z \mapsto H$

    $g$ maps the recurrent states, inputs, and start flags to outputs.

    $g: H^n \times X^n \times \\{0, 1\\}^n \mapsto Y^n$

    We utilize `vmap` and apply our GRAS to some input following the below pseudocode:
    ```
    z = vmap(f)(x, start)
    h = scan(., h0, z)
    y = vmap(g)(h, x, start)
    ```

    Note that we can represent any recurrent function in this form.

    # Efficient GRAS via Semigroups

    A semigroup is a special case of a GRAS where the $\bullet$ is associative

    $ a \bullet (b \bullet c) = (a \bullet b) \bullet c $.

    This enables us to execute $\bullet$ via a parallel scan, which is much more efficient than a sequential scan. 
    The semigroup GRAS therefore contains the same maps $f$ and $g$ as above, but $\bullet$ is now a semigroup operation.
    Furthermore, in a semigroup, the action and recurrent state spaces are identical, i.e., $Z = H$

    $\bullet: H \times H \mapsto H$. We execute semigroup-based GRAS just as above, but using a parallel scan for efficiency
    ```
    z = vmap(f)(x, start)
    h = associative_scan(., h0, z)
    y = vmap(g)(h, x, start)
    ```
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[[RecurrentState, RecurrentState], RecurrentState],
            RecurrentState,
            RecurrentState,
        ],
        RecurrentState,
    ]

    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> RecurrentState:
        """Maps inputs to the recurrent space.
        
        `(feature, start) -> H`
        """
        raise NotImplementedError

    def backward_map(
        self,
        h: RecurrentState,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> OutputEmbedding:
        """Maps the recurrent space to the output space.

        `(h, (feature, start)) -> Y`
        """
        raise NotImplementedError

    @nn.compact
    def __call__(
        self,
        h: SingleRecurrentState,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Tuple[RecurrentState, OutputEmbedding]:
        """Calls the mapping and scan functions.

        ```
        z = vmap(f)(feature, start)
        h = scan(., h0, z)
        y = vmap(g)(h, feature, start)
        ```

        You probably do not need to override this."""
        if key is None:
            in_key, out_key = (None, None)
        else:
            in_key, out_key = jax.random.split(key)
        scan_input = jax.vmap(self.forward_map)(x, in_key)
        next_h = self.scan(self.algebra, h, scan_input)
        y = jax.vmap(self.backward_map)(next_h, x, out_key)
        return next_h, y

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SingleRecurrentState:
        """Initialize the recurrent state for a new sequence."""
        return self.algebra.initialize_carry(key=key)

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        """Linen is "quirky". This function returns a zero-initialized carry/state
        that is often needed before calling initialize_carry in Linen RNNs."""
        return self.algebra.zero_carry()

    @nn.nowrap
    def latest_recurrent_state(self, hs: RecurrentState) -> RecurrentState:
        """Get the latest state from a sequence of recurrent states."""
        return jax.tree.map(lambda x: x[-1], hs)

    @staticmethod
    def default_algebra(**kwargs):
        raise NotImplementedError

    @staticmethod
    def default_scan():
        raise NotImplementedError



from beartype.typing import List, Optional, Tuple

import jax
from equinox import filter_vmap, nn
from jaxtyping import PRNGKeyArray, Shaped

from memax.equinox.groups import Module
from memax.mtypes import Input, ResetRecurrentState


class ResidualModel(Module):
    """A model consisting of multiple recurrent layers, with a
    residual connection from the original input into each layer.

    There is a nonlinearity between network layers."""

    layers: List[Module]
    ff: List[nn.Sequential]
    map_in: nn.Linear
    map_out: nn.Linear

    def __init__(
        self,
        make_layer_fn,
        input_size,
        output_size,
        recurrent_size,
        num_layers=2,
        activation=jax.nn.leaky_relu,
        *,
        key
    ):
        self.layers = []
        self.ff = []
        keys = jax.random.split(key, 3)
        self.map_in = nn.Linear(input_size, recurrent_size, key=keys[0])
        self.map_out = nn.Linear(recurrent_size, output_size, key=keys[1])
        key = keys[2]
        for _ in range(num_layers):
            key, ff_key = jax.random.split(key)
            self.layers.append(make_layer_fn(recurrent_size=recurrent_size, key=key))
            self.ff.append(
                nn.Sequential(
                    [
                        nn.Linear(recurrent_size, recurrent_size, key=ff_key),
                        nn.LayerNorm(
                            (recurrent_size,), use_weight=False, use_bias=False
                        ),
                        nn.Lambda(activation),
                    ]
                )
            )

    def __call__(
        self, h: ResetRecurrentState, x: Input, key: Optional[PRNGKeyArray] = None
    ) -> Tuple[ResetRecurrentState, ...]:
        emb, start = x
        emb = filter_vmap(self.map_in)(emb)
        layer_in = (emb, start)
        h_out = []
        for ff, recurrent_layer, h_i in zip(self.ff, self.layers, h):
            if key is None:
                key, rkey = None, None
            else:
                key, rkey = jax.random.split(key)
            tmp, z = recurrent_layer(h_i, layer_in, key=rkey)
            h_out.append(tmp)
            z = filter_vmap(ff)(z)
            layer_in = (z, start)
        out = filter_vmap(self.map_out)(layer_in[0])
        return tuple(h_out), out

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> Tuple[ResetRecurrentState, ...]:
        if key is None:
            keys = tuple(None for _ in range(len(self.layers)))
        else:
            keys = jax.random.split(key, len(self.layers))
        return tuple(l.initialize_carry(k) for l, k in zip(self.layers, keys))

    def latest_recurrent_state(self, h: ResetRecurrentState) -> ResetRecurrentState:
        return tuple(l.latest_recurrent_state(h_i) for l, h_i in zip(self.layers, h))



"""
Some definitions used in our models
Recurrent layers should take (Input, RecurrentState) 
and return (RecurrentState, OutputEmbedding)
"""
from beartype.typing import Tuple

from jaxtyping import Array, Bool, PyTree, Shaped

# Inputs
StartFlag = Bool[Array, ""]
"""Start flags denote the beginning of a new sequence. They should be
1 where a new sequence starts, and 0 elsewhere."""
InputEmbedding = Shaped[Array, "Feature"]
"""An input embedding are the per-time-step input features to the recurrent model."""
InputEmbeddings = Shaped[Array, "Time Feature"]
"""InputEmbeddings are a temporal sequence of InputEmbedding."""
Input = Tuple[InputEmbedding, StartFlag]
"""All layers take both an input embedding and a start flag as input."""

# Recurrence
RecurrentState = PyTree[Array, "..."]
"""A RecurrentState is the hidden state of a recurrent layer for a single timestep."""
RecurrentStates = PyTree[Array, "Time ..."]
"""RecurrentStates are the hidden states of a recurrent layer for all timesteps."""
SingleRecurrentState = PyTree[Array, "One ..."]
"""A single recurrent state with a leading singleton time dimension."""
ResetRecurrentState = Tuple[RecurrentState, StartFlag]
"""A single recurrent state and reset carry with a leading singleton time dimension."""

# Outputs
OutputEmbedding = Shaped[Array, "..."]
"""The output (not recurrent state) of a recurrent layer for a single timestep."""
OutputEmbeddings = Shaped[Array, "Time ..."]
"""The output (not recurrent state) of a recurrent layer for all timesteps."""



from beartype.typing import Callable, List, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

FWPRecurrentState = Float[Array, "Key Value"]
FWPRecurrentStateWithReset = Tuple[FWPRecurrentState, StartFlag]


def phi(x, key=None):
    return 1 + jax.nn.elu(x)


class FWPSemigroup(Semigroup):
    """The Additive Fast Weight Programmer semigroup (recurrent update) 
    from https://arxiv.org/pdf/2508.08435"""

    recurrent_size: int

    def __init__(self, recurrent_size):
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FWPRecurrentState:
        return jnp.zeros((self.recurrent_size, self.recurrent_size))

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self,
        carry: FWPRecurrentState,
        input: FWPRecurrentState,
    ) -> FWPRecurrentState:
        return carry + input


class FWP(GRAS):
    """The Additive Fast Weight Programmer from https://arxiv.org/pdf/2508.08435

    You might want to use this as a building block for a more complex model.
    """

    hidden_size: int
    recurrent_size: int
    scan: Callable[
        [
            Callable[
                [FWPRecurrentStateWithReset, FWPRecurrentStateWithReset],
                FWPRecurrentStateWithReset,
            ],
            FWPRecurrentStateWithReset,
            FWPRecurrentStateWithReset,
        ],
        FWPRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra

    K: nn.Linear
    Q: nn.Linear
    V: nn.Linear
    output: nn.Linear

    def __init__(self, hidden_size, recurrent_size, key):
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        self.algebra = Resettable(FWPSemigroup(recurrent_size))
        self.scan = semigroup_scan

        keys = jax.random.split(key, 4)

        self.K = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[0])
        self.Q = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[1])
        self.V = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[2])
        self.output = nn.Linear(recurrent_size, hidden_size, key=keys[3])

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FWPRecurrentStateWithReset:
        emb, start = x
        k = phi(self.K(emb))
        v = self.V(emb)
        kv = jnp.outer(k, v)
        return kv, start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: FWPRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        emb, start = x
        kv_sum, reset_flag = h
        q = phi(self.Q(emb))
        return self.output(kv_sum @ q)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FWPRecurrentStateWithReset:
        # inputs should be of shape [*batch, time, feature]
        # recurrent states should be of shape [*batch, 1, feature]
        return self.algebra.initialize_carry(key)



"""Ensure that the reset model is equivalent to a non-reset batched model"""
import equinox as eqx
import jax
import jax.numpy as jnp
import pytest

from memax.equinox.train_utils import add_batch_dim, get_residual_memory_models

@pytest.mark.parametrize("name, model", get_residual_memory_models(
        input=1, hidden=8, output=10, num_layers=2, key=jax.random.key(0)
    ).items()
)
def test_reset(name, model):
    print(f"Testing {name}")
    L = 2048
    B = 32
    x_long = jax.random.uniform(jax.random.key(1), (L, 1))
    x_short = x_long.reshape(B, -1, 1)
    start_long = jnp.zeros(L, dtype=bool).at[jnp.arange(0, L, B)].set(True)
    start_short = start_long.reshape(B, -1)

    h0_long = model.initialize_carry()
    h_long, out_long = model(h0_long, (x_long, start_long))
    h0_short = add_batch_dim(h0_long, B)
    h_short, out_short = eqx.filter_vmap(model)(h0_short, (x_short, start_short))

    assert jnp.allclose(
        out_long,
        out_short.reshape(L, *out_short.shape[2:]),
        atol=1e-5,
        rtol=1e-5,
    ), f"Reset failed for {name}"

if __name__ == "__main__":
    test_reset()


from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, SetAction, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import set_action_scan

GRURecurrentState = Float[Array, "Recurrent"]
GRURecurrentStateWithReset = Tuple[GRURecurrentState, StartFlag]


class GRUMagma(SetAction):
    """
    The Gated Recurrent Unit set action

    Paper: https://arxiv.org/abs/1406.1078
    """

    recurrent_size: int
    U_z: nn.Linear
    U_r: nn.Linear
    U_h: nn.Linear
    W_z: nn.Linear
    W_r: nn.Linear
    W_h: nn.Linear

    def __init__(self, recurrent_size: int, key):
        self.recurrent_size = recurrent_size
        keys = jax.random.split(key, 6)
        self.U_z = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[0]
        )
        self.U_r = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[1]
        )
        self.U_h = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[2]
        )

        self.W_z = nn.Linear(recurrent_size, recurrent_size, key=keys[3])
        self.W_r = nn.Linear(recurrent_size, recurrent_size, key=keys[4])
        self.W_h = nn.Linear(recurrent_size, recurrent_size, key=keys[5])

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: GRURecurrentState, input: Float[Array, "Recurrent"]
    ) -> GRURecurrentState:
        z = jax.nn.sigmoid(self.W_z(input) + self.U_z(carry))
        r = jax.nn.sigmoid(self.W_r(input) + self.U_r(carry))
        h_hat = jax.nn.tanh(self.W_h(input) + self.U_h(r * carry))
        out = (1 - z) * carry + z * h_hat
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentState:
        return jnp.zeros((self.recurrent_size,))


class GRU(GRAS):
    """
    The Gated Recurrent Unit layer

    Paper: https://arxiv.org/abs/1406.1078
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [GRURecurrentStateWithReset, GRURecurrentStateWithReset],
                GRURecurrentStateWithReset,
            ],
            GRURecurrentStateWithReset,
            GRURecurrentStateWithReset,
        ],
        GRURecurrentStateWithReset,
    ]
    recurrent_size: int

    def __init__(self, recurrent_size, key):
        keys = jax.random.split(key, 3)
        self.recurrent_size = recurrent_size
        self.algebra = Resettable(GRUMagma(recurrent_size, key=keys[0]))
        self.scan = set_action_scan

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentStateWithReset:
        emb, start = x
        return emb, start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: GRURecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        z, reset_flag = h
        emb, start = x
        return z

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



"""Proves the correctness (empirically) of associative updates"""
import pytest
from functools import partial

import jax
import jax.numpy as jnp

from memax.equinox.groups import Semigroup
from memax.equinox.train_utils import get_semigroups


def random_state(state, key):
    if state.dtype in [jnp.float32, jnp.complex64]:
        return jax.random.normal(key, state.shape, dtype=state.dtype)
    elif state.dtype in [jnp.int32]:
        return jax.random.randint(key, state.shape, 0, 5, dtype=state.dtype)
    elif state.dtype in [jnp.bool_]:
        return jax.random.bernoulli(key, 0.5, state.shape).astype(state.dtype)
    else:
        raise NotImplementedError(
            f"Random state not implemented for dtype {state.dtype}"
        )


def map_assert(monoid, a, b):
    is_equal = jnp.allclose(a, b)
    if not is_equal:
        raise Exception(
            f"Monoid {type(monoid).__name__} failed associativity test:\n{a} != \n{b}"
        )

def perturb(pytree, key):
    def _perturb(x):
        return (x + jax.random.uniform(key, shape=x.shape)).astype(x.dtype)
    return jax.tree.map(_perturb, pytree)

@pytest.mark.parametrize("name, sg", get_semigroups(recurrent_size=3, key=jax.random.PRNGKey(0)).items())
def test_semigroup_correctness(name: str, sg: Semigroup):
    initial_state = sg.initialize_carry()
    x1 = jax.tree.map(partial(random_state, key=jax.random.key(1)), initial_state)
    x2 = jax.tree.map(partial(random_state, key=jax.random.key(2)), perturb(initial_state, jax.random.key(4)))
    x3 = jax.tree.map(partial(random_state, key=jax.random.key(3)), perturb(initial_state, jax.random.key(5)))

    a = sg(sg(x1, x2), x3)
    b = sg(x1, sg(x2, x3))

    is_equal = jax.tree.map(jnp.allclose, a, b)
    if isinstance(is_equal, tuple):
        is_equal = all(is_equal)
    else:
        is_equal = jnp.all(is_equal)

    jax.tree.map(partial(map_assert, sg), a, b)


if __name__ == '__main__':
    models = get_semigroups(recurrent_size=3, key=jax.random.PRNGKey(0)).items()
    for name, sg in models:
        test_semigroup_correctness(name, sg)



"""Flax Linen backend for memax.

`memax.linen.gras` provides the Generalized Recurrent Algebraic Structure (GRAS) base module.
`memax.linen.groups` provides the algebraic structures (semigroups, groups, etc.) used in GRAS.
`memax.linen.set_actions` provides set action-based recurrent layers (slow RNNs).
`memax.linen.semigroups` provides semigroup-based recurrent layers (fast RNNs).
`memax.linen.scans` provides scan functions for recurrent updates.
"""


from beartype.typing import Callable, List, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
import equinox as eqx
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

GDNRecurrentState = Tuple[
    Float[Array, "Key Value"],
    Float[Array, "Key Value"],
]
GDNRecurrentStateWithReset = Tuple[GDNRecurrentState, StartFlag]


def phi(x, key=None):
    # https://arxiv.org/pdf/2102.11174 uses relu
    # https://arxiv.org/pdf/2406.06484 uses silu
    return jax.nn.silu(x)

def psi(x, key=None):
    # https://arxiv.org/pdf/2102.11174 uses sigmoid
    # https://arxiv.org/pdf/2508.08435 suggests 2 * sigmoid
    return 2 * jax.nn.sigmoid(x)


class GDNSemigroup(Semigroup):
    """The Gated Delta Net semigroup (recurrent update) 
    from https://arxiv.org/pdf/2412.06464"""

    recurrent_size: int

    def __init__(self, recurrent_size):
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GDNRecurrentState:
        return (
            jnp.eye(self.recurrent_size),
            jnp.zeros((self.recurrent_size, self.recurrent_size))
        )

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self,
        carry: GDNRecurrentState,
        input: GDNRecurrentState,
    ) -> GDNRecurrentState:
        # Amazing resource: https://sustcsonglin.github.io/blog/2024/deltanet-2/
        # Based on Songlin's factorization
        M_i, X_i = carry
        M_j, X_j = input
        return M_j @ M_i, M_j @ X_i + X_j


class GDN(GRAS):
    """The Gated Delta Network from https://arxiv.org/pdf/2412.06464

    You might want to use this as a building block for a more complex model.
    """

    hidden_size: int
    recurrent_size: int
    scan: Callable[
        [
            Callable[
                [GDNRecurrentStateWithReset, GDNRecurrentStateWithReset],
                GDNRecurrentStateWithReset,
            ],
            GDNRecurrentStateWithReset,
            GDNRecurrentStateWithReset,
        ],
        GDNRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra

    K: nn.Linear
    Q: nn.Linear
    V: nn.Linear
    w: nn.Linear
    alpha: nn.Linear
    output: nn.Linear

    def __init__(self, hidden_size, recurrent_size, key):
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        self.algebra = Resettable(GDNSemigroup(recurrent_size))
        self.scan = semigroup_scan

        keys = jax.random.split(key, 6)

        self.K = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[0])
        self.Q = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[1])
        self.V = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[2])
        self.w = nn.Linear(hidden_size, 1, key=keys[3])
        alpha = nn.Linear(hidden_size, 1, key=keys[4])
        # Initialize alpha bias to 4.0 so that sigmoid(alpha) is near 1.0 at init
        self.alpha = eqx.tree_at(
            lambda l: l.bias, alpha, jnp.full_like(alpha.bias, 4.0)
        )
        self.output = nn.Linear(recurrent_size, hidden_size, key=keys[5])

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GDNRecurrentStateWithReset:
        emb, start = x
        k = phi(self.K(emb))
        k = k / (jnp.linalg.norm(k) + 1e-6)  # normalize key
        v = self.V(emb)
        beta = psi(self.w(emb))
        alpha = jax.nn.sigmoid(self.alpha(emb))
        M = alpha * (jnp.eye(self.recurrent_size) - beta * jnp.outer(k, k))
        X = beta * jnp.outer(v, k)
        return (M, X), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: GDNRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        emb, start = x
        (M, X), reset_flag = h
        q = phi(self.Q(emb))
        q = q / (jnp.linalg.norm(q) + 1e-6)  # normalize query
        return self.output(X @ q)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GDNRecurrentStateWithReset:
        # inputs should be of shape [*batch, time, feature]
        # recurrent states should be of shape [*batch, 1, feature]
        return self.algebra.initialize_carry(key)



from memax.datasets.continuous_localization import step
import jax.numpy as jnp
from jax.scipy.spatial.transform import Rotation
import jax

def test_step():
    dx = jnp.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
    x = jnp.array([[1, 0, 0], [2, 0, 0], [2, 1, 0]])
    drot = Rotation.from_quat(jnp.stack([
        Rotation.from_euler('z', jnp.pi/2).as_quat(),
        Rotation.from_euler('y', -jnp.pi/2).as_quat(),
        Rotation.from_euler('YZ', (jnp.pi/2, -jnp.pi/2)).as_quat()
    ], axis=0))
    rot = Rotation.from_quat(jnp.stack([
        Rotation.from_euler('z', jnp.pi / 2).as_quat(),
        Rotation.from_euler('zx', (jnp.pi/2, jnp.pi/2)).as_quat(),
        Rotation.identity().as_quat(),
    ], axis=0))

    x_start = jnp.zeros((1,3))
    rot_start = jax.vmap(Rotation.identity, axis_size=1)()

    _, (x_pred, rot_pred) = jax.lax.scan(step, (x_start, rot_start), (dx[:,None], drot[:,None]))

    pred_rot = rot_pred.as_matrix()[:,0]
    true_rot = rot.as_matrix()
    pred_x = x_pred[:,0]
    true_x = x

    assert jnp.allclose(pred_rot, true_rot)
    assert jnp.allclose(pred_x, true_x)


if __name__ == "__main__":
    test_step()


from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
import equinox as eqx
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, SetAction, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import set_action_scan

ElmanRecurrentState = Float[Array, "Recurrent"]
ElmanRecurrentStateWithReset = Tuple[ElmanRecurrentState, StartFlag]


class ElmanSetAction(SetAction):
    """
    The Elman set action

    Paper: https://onlinelibrary.wiley.com/doi/abs/10.1207/s15516709cog1402_1.
    """

    recurrent_size: int
    U_h: nn.Linear
    activation: eqx.Module

    def __init__(self, recurrent_size: int, activation=jax.nn.tanh, *, key):
        self.recurrent_size = recurrent_size
        self.U_h = nn.Linear(recurrent_size, recurrent_size, key=key)
        self.activation = activation

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: ElmanRecurrentState, input: ElmanRecurrentState
    ) -> ElmanRecurrentState:
        return self.activation(self.U_h(carry) + input)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> ElmanRecurrentState:
        return jnp.zeros((self.recurrent_size,))


class Elman(GRAS):
    """
    The Elman Recurrent Network layer

    Paper: https://onlinelibrary.wiley.com/doi/abs/10.1207/s15516709cog1402_1.
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [ElmanRecurrentStateWithReset, ElmanRecurrentStateWithReset],
                ElmanRecurrentStateWithReset,
            ],
            ElmanRecurrentStateWithReset,
            ElmanRecurrentStateWithReset,
        ],
        ElmanRecurrentStateWithReset,
    ]
    recurrent_size: int
    hidden_size: int
    W_h: nn.Linear
    W_y: nn.Linear

    def __init__(self, recurrent_size, hidden_size, activation=jax.nn.tanh, *, key):
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        keys = jax.random.split(key, 3)
        self.algebra = Resettable(ElmanSetAction(recurrent_size, activation=activation, key=keys[0]))
        self.scan = set_action_scan
        self.W_h = nn.Linear(hidden_size, recurrent_size, use_bias=False, key=keys[1])
        self.W_y = nn.Linear(recurrent_size, hidden_size, key=keys[2])

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> ElmanRecurrentStateWithReset:
        emb, start = x
        return self.W_h(emb), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: ElmanRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        z, reset_flag = h
        emb, start = x
        return self.W_y(z)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> ElmanRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, SetAction, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, InputEmbedding, StartFlag
from memax.equinox.scans import set_action_scan

LSTMRecurrentState = Tuple[Float[Array, "Recurrent"], Float[Array, "Recurrent"]]
LSTMRecurrentStateWithReset = Tuple[LSTMRecurrentState, StartFlag]


class LSTMMagma(SetAction):
    """
    The Long Short-Term Memory set action

    Paper: https://www.bioinf.jku.at/publications/older/2604.pdf
    """

    recurrent_size: int
    U_f: nn.Linear
    U_i: nn.Linear
    U_o: nn.Linear
    U_c: nn.Linear
    W_f: nn.Linear
    W_i: nn.Linear
    W_o: nn.Linear
    W_c: nn.Linear

    def __init__(self, recurrent_size: int, key):
        self.recurrent_size = recurrent_size
        keys = jax.random.split(key, 8)
        self.U_f = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[0]
        )
        self.U_i = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[1]
        )
        self.U_o = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[2]
        )
        self.U_c = nn.Linear(
            recurrent_size, recurrent_size, use_bias=False, key=keys[3]
        )

        self.W_f = nn.Linear(recurrent_size, recurrent_size, key=keys[4])
        self.W_i = nn.Linear(recurrent_size, recurrent_size, key=keys[5])
        self.W_o = nn.Linear(recurrent_size, recurrent_size, key=keys[6])
        self.W_c = nn.Linear(recurrent_size, recurrent_size, key=keys[7])

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: LSTMRecurrentState, input: LSTMRecurrentState
    ) -> LSTMRecurrentState:
        x_t, _ = input
        c, h = carry
        f_f = jax.nn.sigmoid(self.W_f(x_t) + self.U_f(h))
        f_i = jax.nn.sigmoid(self.W_i(x_t) + self.U_i(h))
        f_o = jax.nn.sigmoid(self.W_o(x_t) + self.U_o(h))
        f_c = jax.nn.tanh(self.W_c(x_t) + self.U_c(h))

        c = f_f * c + f_i * f_c
        h = f_o * c

        return (c, h)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LSTMRecurrentState:
        return (
            jnp.zeros((self.recurrent_size,)),
            jnp.zeros((self.recurrent_size,)),
        )


class LSTM(GRAS):
    """
    The Long Short-Term Memory layer

    Paper: https://www.bioinf.jku.at/publications/older/2604.pdf
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [LSTMRecurrentStateWithReset, LSTMRecurrentStateWithReset],
                LSTMRecurrentStateWithReset,
            ],
            LSTMRecurrentStateWithReset,
            LSTMRecurrentStateWithReset,
        ],
        LSTMRecurrentStateWithReset,
    ]

    def __init__(self, recurrent_size, key):
        keys = jax.random.split(key, 3)
        self.algebra = Resettable(LSTMMagma(recurrent_size, key=keys[0]))
        self.scan = set_action_scan

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LSTMRecurrentStateWithReset:
        emb, start = x
        c = jnp.zeros_like(emb)
        return (emb, c), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: LSTMRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) ->  Float[Array, "Recurrent"]:
        (c_t, h_t), reset_flag = h
        emb, start = x
        return h_t

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LSTMRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped, Bool, Int

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan
from memax.utils import apply_rope, apply_sinusoidal_pe, combine_and_right_align

AttentionRecurrentState = Tuple[Float[Array, "Window Recurrent"], Float[Array, "Window Recurrent"], Bool[Array, "Window"], Int[Array, "Window"]]
AttentionRecurrentStateWithReset = Tuple[AttentionRecurrentState, StartFlag]


class AttentionSemigroup(Semigroup):
    """A sliding window attention semigroup example.

    See the Stack semigroup for how to implement sliding windows.
    """
    window_size: int
    recurrent_size: int

    def __init__(self, recurrent_size: int, window_size: int):
        self.window_size = window_size 
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> AttentionRecurrentState:
        key = jnp.zeros((self.window_size, self.recurrent_size))
        value = jnp.zeros((self.window_size, self.recurrent_size))
        # Valid (non-pad) mask
        mask = jnp.zeros((self.window_size,), dtype=bool)
        ts = jnp.zeros((self.window_size), dtype=jnp.int32)
        return (key, value, mask, ts)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: AttentionRecurrentState, input: AttentionRecurrentState
    ) -> AttentionRecurrentState:
        # We would like to do the below
        # But cannot due to concretization error
        # Caused by dynamic indexing (mask)
        #
        # left = cwindow[cmask]
        # right = window[mask]

        # mleft = cmask[cmask]
        # mright = mask[mask]

        # window = jnp.concatenate([left, right])[-window_size:]
        # mask = jnp.concatenate([mleft, mright])[-window_size:]

        # So we use a tricky function instead
        ckey, cvalue, cmask, cts = carry
        key, value, mask, ts = input
        out_key, out_mask = combine_and_right_align(ckey, cmask, key, mask)
        out_value, _ = combine_and_right_align(cvalue, cmask, value, mask)
        out_ts = cts + ts
        return (out_key, out_value, out_mask, out_ts)


class Attention(GRAS):
    """Standard dot-product attention with a sliding window. This utilizes
    the Stack semigroup for maintaining a recurrent sliding window.
    """

    K: nn.Linear
    Q: nn.Linear
    V: nn.Linear
    recurrent_size: int
    window_size: int
    positional_embedding: Optional[str]
    scan: Callable[
        [
            Callable[
                [AttentionRecurrentStateWithReset, AttentionRecurrentStateWithReset],
                AttentionRecurrentStateWithReset,
            ],
            AttentionRecurrentStateWithReset,
            AttentionRecurrentStateWithReset,
        ],
        AttentionRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra


    def __init__(self, recurrent_size: int, window_size: int, positional_embedding: Optional[str], key):
        """Standard dot-product attention with a sliding window. 
        Arguments:
            recurrent_size: The size of the attention embeddings.
            window_size: The size of the attention window (context length).
            rope: Whether to use RoPE embeddings (False means no embeddings).
        """
        assert positional_embedding in [None, "rope", "alibi"], "positional_embedding must be one of None, 'rope', or 'alibi'"
        self.recurrent_size = recurrent_size
        self.window_size = window_size
        self.positional_embedding = positional_embedding
        self.algebra = Resettable(AttentionSemigroup(recurrent_size, window_size=window_size))
        self.scan = semigroup_scan
        keys = jax.random.split(key, 5)
        self.K = nn.Linear(recurrent_size, recurrent_size, use_bias=False, key=keys[0])
        self.Q = nn.Linear(recurrent_size, recurrent_size, use_bias=False, key=keys[1])
        self.V = nn.Linear(recurrent_size, recurrent_size, key=keys[2])

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> AttentionRecurrentStateWithReset:
        emb, start = x
        # Add Attention dim for concat
        mask = jnp.concatenate([
            jnp.zeros((self.window_size - 1), dtype=bool),
            jnp.ones((1,), dtype=bool)
        ])
        k = self.K(emb)
        v = self.V(emb)
        key = jnp.concatenate([
            jnp.zeros((self.window_size - 1, *emb.shape), dtype=emb.dtype),
            k.reshape(1, -1)
        ])
        value = jnp.concatenate([
            jnp.zeros((self.window_size - 1, *emb.shape), dtype=emb.dtype),
            v.reshape(1, -1)
        ])
        ts = jnp.ones((self.window_size), dtype=jnp.int32)
        return (key, value, mask, ts), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: AttentionRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        emb, start = x
        state, reset_carry = h
        K, V, mask, ts = state
        q = self.Q(emb)

        # B = batch size
        # S = length of the key/value (source)
        # T = length of the query (target)
        # N = number of attention heads
        # H = dimensions of each attention head
        # K = number of key/value heads
        # G = number of groups, which equals to N // K
        n, k, t, s, h = 1, 1, 1, self.window_size, self.recurrent_size
        bias = None
        if self.positional_embedding == "alibi":
            m = 2 ** -8
            # T-1 to 0
            bias = m * (ts[0] + jnp.arange(-s + 1, 1))
        elif self.positional_embedding == "rope":
            K, q = apply_rope(K, q)

        mask = mask.reshape(n, t, s)
        bias = bias if bias is None else bias.reshape(n, t, s)
        K = K.reshape(s, k, h)
        q = q.reshape(t, n, h) # Only for current timestep
        V = V.reshape(s, k, h)
        z = jax.nn.dot_product_attention(q, K, V, mask=mask, bias=bias)
        return z.reshape(-1)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> AttentionRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, List, Optional, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

RotationMatrix = Float[Array, "Hidden Hidden"]
SphericalRecurrentState = RotationMatrix
SphericalRecurrentStateWithReset = Tuple[SphericalRecurrentState, StartFlag]


class PSphericalSemigroup(Semigroup):
    recurrent_size: int

    def __init__(self, recurrent_size: int):
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentState:
        return jnp.eye(self.recurrent_size)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: SphericalRecurrentState, input: SphericalRecurrentState
    ) -> SphericalRecurrentState:
        return carry @ input


class PSpherical(GRAS):
    """A simple Bayesian memory model.

    You might want to use this as a building block for a more complex model.
    """

    recurrent_size: int
    hidden_size: int
    proj_size: int
    scan: Callable[
        [
            Callable[
                [SphericalRecurrentStateWithReset, SphericalRecurrentStateWithReset],
                SphericalRecurrentStateWithReset,
            ],
            SphericalRecurrentStateWithReset,
            SphericalRecurrentStateWithReset,
        ],
        SphericalRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra
    initial_vector: jax.Array
    project: nn.Linear
    output: nn.Linear

    def __init__(self, recurrent_size, hidden_size, key):
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        keys = jax.random.split(key)
        self.proj_size = int(self.recurrent_size * (self.recurrent_size - 1) / 2)
        self.project = nn.Linear(hidden_size, self.proj_size, key=keys[0])
        self.algebra = Resettable(PSphericalSemigroup(recurrent_size))
        self.output = nn.Linear(recurrent_size, hidden_size, key=keys[1])
        self.scan = semigroup_scan
        self.initial_vector = jnp.ones(self.recurrent_size)

    @jaxtyped(typechecker=typechecker)
    def rot(self, x) -> RotationMatrix:
        q = self.project(x)
        A = jnp.zeros((self.recurrent_size, self.recurrent_size))
        tri_idx = jnp.triu_indices_from(A, 1)
        A = A.at[tri_idx].set(q)
        A = A - A.T
        R = jax.scipy.linalg.expm(A)
        return R

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentStateWithReset:
        emb, start = x
        rot = self.rot(emb)
        return rot, start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: SphericalRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        emb, start = x
        state, reset_carry = h

        normed = self.initial_vector / jnp.linalg.norm(self.initial_vector)
        return self.output(state @ normed)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, Optional, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Complex, Float, Int, PRNGKeyArray, Real, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

FFMRecurrentState = Tuple[Complex[Array, "Trace Context"], Int[Array, ""]]
FFMRecurrentStateWithReset = Tuple[FFMRecurrentState, StartFlag]


class Gate(eqx.Module):
    linear: nn.Linear

    def __init__(self, input_size, output_size, key):
        self.linear = nn.Linear(input_size, output_size, key=key)

    def __call__(self, x):
        return jax.nn.sigmoid(self.linear(x))


class FFMSemigroup(Semigroup):
    """The Fast and Forgetful Memory semigroup (recurrent update) from https://arxiv.org/abs/2310.04128."""

    trace_size: int
    context_size: int
    params: Tuple[Float[Array, "Trace"], Float[Array, "Context"]]

    def __init__(
        self,
        trace_size,
        context_size,
        deterministic_init,
        min_period=1,
        max_period=1024,
        *,
        key,
    ):
        self.trace_size = trace_size
        self.context_size = context_size
        if deterministic_init:
            a_low = 1e-6
            a_high = 0.5
            a = jnp.linspace(a_low, a_high, trace_size)
            b = 2 * jnp.pi / jnp.linspace(min_period, max_period, context_size)
            self.params = (a, b)
        else:
            raise NotImplementedError

    @jaxtyped(typechecker=typechecker)
    def log_gamma(self, t: Real[Array, ""]) -> Complex[Array, "Trace Context"]:
        a, b = self.params
        a = -jnp.abs(a).reshape((self.trace_size, 1))
        b = b.reshape(1, self.context_size)
        ab = jax.lax.complex(a, b)
        return ab * t.reshape(1, 1)

    @jaxtyped(typechecker=typechecker)
    def gamma(self, t: Real[Array, ""]) -> Complex[Array, "Trace Context"]:
        return jnp.exp(self.log_gamma(t))

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FFMRecurrentState:
        # inputs should be of shape [*batch, time, feature]
        # recurrent states should be of shape [*batch, 1, feature]
        carry_shape = (self.trace_size, self.context_size)

        return jnp.zeros(carry_shape, dtype=jnp.complex64), jnp.array(
            0, dtype=jnp.int32
        )

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: FFMRecurrentState, input: FFMRecurrentState
    ) -> FFMRecurrentState:
        (
            state,
            i,
        ) = carry
        x, j = input
        state = state * self.gamma(j) + x
        return state, j + i


class FFM(GRAS):
    """Fast and Forgetful Memory from https://arxiv.org/abs/2310.04128."""

    hidden_size: int
    trace_size: int
    context_size: int
    scan: Callable[
        [
            Callable[
                [FFMRecurrentStateWithReset, FFMRecurrentStateWithReset],
                FFMRecurrentStateWithReset,
            ],
            FFMRecurrentStateWithReset,
            FFMRecurrentStateWithReset,
        ],
        FFMRecurrentStateWithReset,
    ]
    algebra: BinaryAlgebra

    pre: nn.Linear
    gate_in: Gate
    gate_out: Gate
    mix: nn.Linear
    ln: nn.LayerNorm

    def __init__(
        self,
        hidden_size: int,
        trace_size: int,
        context_size: int,
        key: Array,
    ):
        self.hidden_size = hidden_size
        self.trace_size = trace_size
        self.context_size = context_size
        self.scan = semigroup_scan

        k1, k2, k3, k4, k5, k6 = jax.random.split(key, 6)
        self.pre = nn.Linear(hidden_size, trace_size, key=k1)
        self.gate_in = Gate(hidden_size, trace_size, key=k2)
        self.gate_out = Gate(hidden_size, hidden_size, key=k3)
        self.algebra = Resettable(
            FFMSemigroup(trace_size, context_size, True, 1, 10_000, key=k4)
        )
        self.mix = nn.Linear(2 * trace_size * context_size, hidden_size, key=k5)
        self.ln = nn.LayerNorm(hidden_size, use_weight=False, use_bias=False)

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FFMRecurrentStateWithReset:
        emb, start = x
        gate_in = self.gate_in(emb)
        pre = self.pre(emb)
        gated = pre * gate_in
        scan_input = jnp.repeat(
            jnp.expand_dims(gated, 1), self.context_size, axis=1
        ).astype(jnp.complex64)
        dt = jnp.array(1)
        return (scan_input, dt), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: FFMRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        (z, dt), reset_flag = h
        emb, start = x
        z = jnp.concatenate([jnp.real(z), jnp.imag(z)], axis=-1).reshape(-1)
        z = self.mix(z)
        gate_out = self.gate_out(emb)
        out = self.ln(z * gate_out) + emb * (1 - gate_out)
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FFMRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



"""
This module contains framework-agnostic utility functions used throughout the memax library.
"""

from typing import Tuple
import jax
import jax.numpy as jnp
from jaxtyping import Array, Int, Shaped, Float


def debug_shape(x: jax.Array) -> str:
    """Pretty-prints the structure of an arbitrary pytree and the shape
    and dtype of its array nodes."""
    import equinox as eqx

    return eqx.tree_pprint(jax.tree.map(lambda x: {x.shape: x.dtype}, x))

def apply_rope(keys: Float[Array, "Time Feat"], query: Float[Array, "Feat"]) -> Tuple[Float[Array, "Time Feat"], Float[Array, "Feat"]]:
    """
    Applies RoPE assuming contiguous time indices.
    
    Constraints:
    - Keys correspond to time steps [1, 2, ..., T]
    - Query corresponds to time step T
    
    Args:
        keys: Array of shape (T, F)
        query: Array of shape (F)
        
    Returns:
        keys_rope: Embedded keys (T, F)
        query_rope: Embedded query (F)
    """
    T, F = keys.shape
    assert F % 2 == 0, "Feature dimension must be even"
    
    # 1. Generate the Position Indices based on shape T
    # Keys are positions 1 to T
    key_indices = jnp.arange(1, T + 1, dtype=jnp.float32) 
    # Query is position T
    query_index = jnp.array(T, dtype=jnp.float32)

    # 2. Calculate RoPE Frequencies (Theta)
    # Standard formula: theta_i = 10000^(-2i/d)
    theta_indices = jnp.arange(0, F, 2)
    theta = 1.0 / (10000.0 ** (theta_indices / F)) # Shape: (F/2,)

    # 3. Create Complex Rotation Angles
    # Keys: (T, F/2) -> broadcast positions against frequencies
    key_angles = jnp.outer(key_indices, theta)
    # Query: (F/2,) -> scalar T against frequencies
    query_angle = query_index * theta

    # Calculate rotation vectors: e^(i * angle)
    key_rotators = jnp.exp(1j * key_angles)
    query_rotator = jnp.exp(1j * query_angle)

    # 4. Apply Rotation using Complex Numbers
    # Reshape (T, F) -> (T, F/2, 2) and convert to complex
    keys_complex = keys.reshape(T, -1, 2)
    keys_complex = keys_complex[..., 0] + 1j * keys_complex[..., 1]
    
    # Reshape (F) -> (F/2, 2) and convert to complex
    query_complex = query.reshape(-1, 2)
    query_complex = query_complex[..., 0] + 1j * query_complex[..., 1]

    # Multiply (rotate)
    keys_out_complex = keys_complex * key_rotators
    query_out_complex = query_complex * query_rotator

    # 5. Convert back to Real
    keys_rope = jnp.stack([keys_out_complex.real, keys_out_complex.imag], axis=-1).reshape(T, F)
    query_rope = jnp.stack([query_out_complex.real, query_out_complex.imag], axis=-1).reshape(F)

    return keys_rope, query_rope

def apply_sinusoidal_pe(keys: Float[Array, "Time Feat"], query: Float[Array, "Feat"], offset: Int[Array, ""] = jnp.array(0)):
    """
    Applies Standard Sinusoidal Positional Encoding with a temporal offset.
    
    Args:
        keys: Array of shape (T, F).
        query: Array of shape (F).
        offset: (int or scalar) The starting time offset. 
                If offset=10, keys map to positions 11...10+T.
        
    Returns:
        keys_pe: keys + PE(pos)
        query_pe: query + PE(pos)
    """
    T, F = keys.shape
    # Don't allow python ints which force recompile
    assert isinstance(offset, jax.Array), "Offset must be a JAX array scalar."
    assert F % 2 == 0, "Feature dimension F must be even."

    # 1. Define Positions with Offset
    # Cast to float32 immediately for instruction efficiency in sin/cos later
    offset_arr = jnp.array(offset, dtype=jnp.float32)
    
    # Keys: [1+offset, 2+offset, ..., T+offset]
    key_positions = jnp.arange(1, T + 1, dtype=jnp.float32) + offset_arr
    key_positions = key_positions[:, None] # Shape (T, 1) for broadcasting
    
    # Query: T + offset
    # corresponds to the last time step in this batch
    query_position = (jnp.array(T, dtype=jnp.float32) + offset_arr)

    # 2. Calculate Frequency Divisor
    # Note: Standard simplified implementation is just F. 
    # The exact Vaswani paper uses exp(arange(0, d, 2) * -(log(10000.0) / d))
    
    dim_indices = jnp.arange(0, F, 2, dtype=jnp.float32)
    div_term = jnp.exp(dim_indices * -(jnp.log(10000.0) / F)) # Shape (F/2,)
    
    # 3. Generate Embeddings for Keys
    # Broadcast (T, 1) * (F/2,) -> (T, F/2)
    k_args = key_positions * div_term
    
    # Interleave Sin/Cos for keys
    # Shape: (T, F/2, 2) -> Flatten to (T, F)
    pe_keys = jnp.stack([jnp.sin(k_args), jnp.cos(k_args)], axis=-1).reshape(T, F)
    
    # 4. Generate Embeddings for Query
    # Broadcast Scalar * (F/2,) -> (F/2,)
    q_args = query_position * div_term
    
    # Interleave Sin/Cos for query
    pe_query = jnp.stack([jnp.sin(q_args), jnp.cos(q_args)], axis=-1).reshape(F)

    # 5. Add to Inputs
    return keys + pe_keys, query + pe_query

def combine_and_right_align(
    left_array: Shaped[Array, "Time Feat"],
    left_mask: Shaped[Array, "Time"],
    right_array: Shaped[Array, "Time Feat"],
    right_mask: Shaped[Array, "Time"],
) -> Tuple[Shaped[Array, "Time Feat"], Shaped[Array, "Time"]]:
    """
    JIT-compatible function to combine two masked arrays and right-align the result.

    This is the JAX-native equivalent of:

    left = left_array[left_mask]
    right = right_array[right_mask]
    combined = jnp.concatenate([left, right])
    # Manual right-alignment:
    result = jnp.zeros_like(left_array)
    result = result.at[-len(combined):].set(combined)
    """
    # Ensure inputs are JAX arrays and masks are boolean
    left_mask = left_mask.astype(bool)
    right_mask = right_mask.astype(bool)

    # N is the static size of the output array.
    N = left_array.shape[0]

    # 1. Count the number of valid items in each input.
    num_valid_left = jnp.sum(left_mask)
    num_valid_right = jnp.sum(right_mask)

    # 2. Allocate slots based on priority.
    #    The right array gets priority, but can't take more than N slots.
    num_to_keep_right = jnp.minimum(num_valid_right, N)
    #    The left array gets the remaining space.
    space_for_left = N - num_to_keep_right
    num_to_keep_left = jnp.minimum(num_valid_left, space_for_left)

    # 3. Create new, "truncated" masks that select the rightmost valid items.
    #    We do this by ranking the valid items from the right using a flipped cumsum.
    right_ranks_left = jnp.flip(jnp.cumsum(jnp.flip(left_mask)))
    new_left_mask = (right_ranks_left <= num_to_keep_left) & left_mask

    right_ranks_right = jnp.flip(jnp.cumsum(jnp.flip(right_mask)))
    new_right_mask = (right_ranks_right <= num_to_keep_right) & right_mask

    # 4. Combine the arrays and their NEW truncated masks.
    combined_stack = jnp.concatenate([left_array, right_array])
    combined_mask = jnp.concatenate([new_left_mask, new_right_mask])

    # 5. Calculate destination indices for right-alignment.
    total_valid = num_to_keep_left + num_to_keep_right
    start_index = N - total_valid
    ranks = jnp.cumsum(combined_mask) - 1
    dest_indices = start_index + ranks

    # 6. Perform the standard scatter operation.
    new_stack = jnp.zeros_like(left_array)
    updates = jnp.where(combined_mask.reshape(-1, 1), combined_stack, 0)
    new_stack = new_stack.at[dest_indices].add(updates)

    # 7. Generate the final right-aligned mask.
    new_mask = jnp.arange(N) >= start_index

    return new_stack, new_mask



# https://github.com/NicolasZucchet/minimal-LRU/blob/main/lru/model.py
from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Complex, Float, PRNGKeyArray, Scalar, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

LRURecurrentState = Tuple[Complex[Array, "Recurrent"], Complex[Array, "Recurrent"]]
LRURecurrentStateWithReset = Tuple[LRURecurrentState, StartFlag]


# Inits
@jaxtyped(typechecker=typechecker)
def glorot_init(
    key: PRNGKeyArray, shape: Tuple[int, ...], normalization: Scalar = jnp.array(1.0)
):
    return jax.random.normal(key=key, shape=shape) / normalization


class LRUSemigroup(Semigroup):
    """The Linear Recurrent Unit semigroup (recurrent update) from https://arxiv.org/abs/2303.06349."""

    recurrent_size: int

    def __init__(
        self,
        recurrent_size: int,
    ):
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LRURecurrentState:
        # Represent a diagonal matrix as a vector
        return (
            jnp.ones((self.recurrent_size,), dtype=jnp.complex64),
            jnp.zeros((self.recurrent_size), dtype=jnp.complex64),
        )

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: LRURecurrentState, input: LRURecurrentState
    ) -> LRURecurrentState:
        # Ax + Bu, but A is diagonal, and we treat it as a vector
        # So we can be more efficient by writing Ax as vec(A) * x
        A_i, bu_i = carry
        A_j, bu_j = input
        return A_j * A_i, A_j * bu_i + bu_j


class LRU(GRAS):
    """
    The Linear Recurrent Unit from https://arxiv.org/abs/2303.06349.

    You might want to use this as a building block for a more complex model.
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [LRURecurrentStateWithReset, LRURecurrentStateWithReset],
                LRURecurrentStateWithReset,
            ],
            LRURecurrentStateWithReset,
            LRURecurrentStateWithReset,
        ],
        LRURecurrentStateWithReset,
    ]
    gamma_log: Float[Array, "Recurrent"]
    B_re: Float[Array, "Recurrent Hidden"]
    B_im: Float[Array, "Recurrent Hidden"]
    C_re: Float[Array, "Hidden Recurrent"]
    C_im: Float[Array, "Hidden Recurrent"]
    D: Float[Array, "Hidden"]
    nu_log: Float[Array, "Recurrent"]
    theta_log: Float[Array, "Recurrent"]
    gamma_log: Float[Array, "Recurrent"]

    hidden_size: int  # input and output dimensions
    recurrent_size: int  # hidden state dimension
    r_min: float = 0.0
    r_max: float = 1.0
    max_phase: float = jnp.pi * 2

    def __init__(self, recurrent_size, hidden_size, key):
        keys = jax.random.split(key, 7)
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        unwrapped = LRUSemigroup(recurrent_size)
        self.algebra = Resettable(unwrapped)
        self.scan = semigroup_scan

        u1 = jax.random.uniform(keys[5], (self.recurrent_size,))
        u2 = jax.random.uniform(keys[6], (self.recurrent_size,))
        self.nu_log = jnp.log(
            -0.5 * jnp.log(u1 * (self.r_max**2 - self.r_min**2) + self.r_min**2)
        )
        self.theta_log = jnp.log(self.max_phase * u2)

        self.B_re = glorot_init(
            keys[0],
            (self.recurrent_size, self.hidden_size),
            normalization=jnp.sqrt(2 * self.hidden_size),
        )
        self.B_im = glorot_init(
            keys[1],
            (self.recurrent_size, self.hidden_size),
            normalization=jnp.sqrt(2 * self.hidden_size),
        )
        self.C_re = glorot_init(
            keys[2],
            (self.hidden_size, self.recurrent_size),
            normalization=jnp.sqrt(self.recurrent_size),
        )
        self.C_im = glorot_init(
            keys[3],
            (self.hidden_size, self.recurrent_size),
            normalization=jnp.sqrt(self.recurrent_size),
        )
        self.D = glorot_init(keys[4], (self.hidden_size,))

        self.gamma_log = jnp.log(jnp.sqrt(1 - jnp.abs(self.diag_lambda()) ** 2))

    @jaxtyped(typechecker=typechecker)
    def diag_lambda(self) -> Complex[Array, "Recurrent"]:
        return jnp.exp(-jnp.exp(self.nu_log) + 1j * jnp.exp(self.theta_log))

    @jaxtyped(typechecker=typechecker)
    def forward_map(self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None):
        emb, start = x
        B_norm = jax.lax.complex(self.B_re, self.B_im) * jnp.expand_dims(
            jnp.exp(self.gamma_log), axis=-1
        )
        Bu = B_norm @ emb.astype(jnp.complex64)
        return (self.diag_lambda(), Bu), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: LRURecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        state, reset_flag = h
        emb, start = x
        lambdas, lambda_x_Bu = state
        C = jax.lax.complex(self.C_re, self.C_im)
        out = (C @ lambda_x_Bu).real + self.D * emb
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> LRURecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from beartype.typing import Callable, Optional, Tuple

import flax.linen as nn
import jax
from jaxtyping import PRNGKeyArray, Shaped

from memax.mtypes import Input, ResetRecurrentState
from memax.linen.groups import Module


class ResidualModel(Module):
    """A model consisting of multiple recurrent layers, with a
    residual connection from the original input into each layer.

    There is a nonlinearity between network layers."""

    make_layer_fn: Callable[..., nn.Module]
    output_size: int
    recurrent_size: int
    num_layers: int = 2
    activation: Callable[[jax.Array], jax.Array] = jax.nn.leaky_relu

    def setup(self):
        layers = []
        ff = []
        self.map_in = nn.Dense(self.recurrent_size)
        self.map_out = nn.Dense(self.output_size)
        for _ in range(self.num_layers):
            layers.append(self.make_layer_fn(recurrent_size=self.recurrent_size))
            ff.append(
                nn.Sequential(
                    [
                        nn.Dense(self.recurrent_size),
                        nn.LayerNorm(use_scale=False, use_bias=False),
                        self.activation,
                    ]
                )
            )
        self.layers = layers
        self.ff = ff

    def __call__(
        self, h: ResetRecurrentState, x: Input
    ) -> Tuple[ResetRecurrentState, ...]:
        emb, start = x
        emb = jax.vmap(self.map_in)(emb)
        layer_in = (emb, start)
        h_out = []
        for ff, recurrent_layer, h_i in zip(self.ff, self.layers, h):
            tmp, z = recurrent_layer(h_i, layer_in)
            h_out.append(tmp)
            z = jax.vmap(ff)(z)
            layer_in = (z, start)
        out = jax.vmap(self.map_out)(layer_in[0])
        return tuple(h_out), out

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> Tuple[ResetRecurrentState, ...]:
        if key is None:
            keys = tuple(None for _ in range(self.num_layers))
        else:
            keys = jax.random.split(key, self.num_layers)

        return tuple(l.initialize_carry(k) for l, k in zip(self.layers, keys))

    @nn.nowrap
    def zero_carry(self) -> Tuple[ResetRecurrentState, ...]:
        layers = [
            self.make_layer_fn(recurrent_size=self.recurrent_size)
            for _ in range(self.num_layers)
        ]
        return tuple(l.zero_carry() for l in layers)

    @nn.nowrap
    def latest_recurrent_state(self, h: ResetRecurrentState) -> ResetRecurrentState:
        layers = [
            self.make_layer_fn(recurrent_size=self.recurrent_size)
            for _ in range(self.num_layers)
        ]
        return tuple(l.latest_recurrent_state(h_i) for l, h_i in zip(layers, h))




"""This module contains set-action-based (classical) recurrent layers.
Each RNN type gets its own file.
+ `memax.equinox.set_actions.elman` provides a basic Elman RNN layer.
+ `memax.equinox.set_actions.lstm` provides the Long Short-Term Memory layer.
+ `memax.equinox.set_actions.gru` provides the Gated Recurrent Unit layer.
+ `memax.equinox.set_actions.mru` provides the Minimal Gated Unit layer
+ `memax.equinox.set_actions.spherical` provides a recurrent formulation of the Rotational RNN. See `memax.equinox.semigroups.spherical` for the semigroup version.
"""


"""This module contains training utilities for Equinox models.
It includes loss functions, accuracy metrics, and training loops.
It also provides a straightforward way to construct multi-layer recurrent models."""

from beartype.typing import Callable, Dict, Tuple, Any, Optional

import equinox as eqx
import jax
import jax.numpy as jnp
import optax
from jaxtyping import Array, Shaped

from memax.equinox.groups import Module
from memax.equinox.set_actions.elman import Elman
from memax.equinox.set_actions.gru import GRU
from memax.equinox.set_actions.lstm import LSTM
from memax.equinox.set_actions.mgu import MGU
from memax.equinox.set_actions.spherical import Spherical
from memax.equinox.models.residual import ResidualModel
from memax.equinox.semigroups.fwp import FWP, FWPSemigroup
from memax.equinox.semigroups.fart import FART, FARTSemigroup
from memax.equinox.semigroups.ffm import FFM, FFMSemigroup
from memax.equinox.semigroups.lrnn import LinearRecurrent, LinearRNNSemigroup
from memax.equinox.semigroups.lru import LRU, LRUSemigroup
from memax.equinox.semigroups.nmax import NMax, NMaxSemigroup
from memax.equinox.semigroups.spherical import PSpherical, PSphericalSemigroup
from memax.equinox.semigroups.s6 import S6, S6Semigroup
from memax.equinox.semigroups.delta import DeltaNet, DeltaNetSemigroup
from memax.equinox.semigroups.deltap import DeltaProduct, DeltaProductSemigroup
from memax.equinox.semigroups.gdn import GDN, GDNSemigroup
from memax.equinox.semigroups.mlp import MLP
from memax.equinox.semigroups.stack import Stack, StackSemigroup
from memax.equinox.semigroups.attn import Attention, AttentionSemigroup


def add_batch_dim(h, batch_size: int, axis: int = 0) -> Shaped[Array, "Batch ..."]:
    """Given an recurrent state (pytree) `h`, add a new batch dimension of size `batch_size`.

    E.g., add_batch_dim(h, 32) will return a new state with shape (32, *h.shape). The state will
    be repeated along the new batch dimension.
    """
    expand = lambda x: jnp.repeat(jnp.expand_dims(x, axis), batch_size, axis=axis)
    h = jax.tree.map(expand, h)
    return h

def cross_entropy(
    y_hat: Shaped[Array, "Batch ... Classes"], y: Shaped[Array, "Batch ... Classes"]
) -> Shaped[Array, "1"]:
    return -jnp.mean(jnp.sum(y * jax.nn.log_softmax(y_hat, axis=-1), axis=-1))

def mse(
    y_hat: Shaped[Array, "Batch ... Feature"], y: Shaped[Array, "Batch ... Feature"]
) -> Shaped[Array, "1"]:
    return jnp.mean(jnp.linalg.norm(y - y_hat, axis=-1, ord=2))

def l1_error(
    y_hat: Shaped[Array, "Batch ... Feature"], y: Shaped[Array, "Batch ... Feature"]
) -> Shaped[Array, "1"]:
    return jnp.mean(jnp.linalg.norm(y - y_hat, axis=-1, ord=1))

def accuracy(
    y_hat: Shaped[Array, "Batch ... Classes"], y: Shaped[Array, "Batch ... Classes"]
) -> Shaped[Array, "1"]:
    return jnp.mean(jnp.argmax(y, axis=-1) == jnp.argmax(y_hat, axis=-1))

def loss_regress_terminal_output(
    model: Module,
    x: Shaped[Array, "Batch Time Feature"],
    y: Shaped[Array, "Batch Classes"],
    key = None
) -> Tuple[Shaped[Array, "1"], Dict[str, Array]]:
    """Given a sequence of inputs x1, ..., xn and predicted outputs y1p, ..., y1n,
    return the mean square error loss between the true yn and predicted y1n.

    Args:
        model: Module
        x: (batch, time, in_feature)
        y: (batch, out_feature)

    Returns:
        loss: scalar
        info: dict
    """
    batch_size = x.shape[0]
    seq_len = x.shape[1]
    assert (
        x.shape[0] == y.shape[0]
    ), f"batch size mismatch: {x.shape[0]} != {y.shape[0]}"
    assert x.ndim == 3, f"expected 3d input, got {x.ndim}d"
    assert y.ndim == 2, f"expected 2d input, got {y.ndim}d"

    starts = jnp.zeros((batch_size, seq_len), dtype=bool)
    key, init_key, model_key = jax.random.split(key, 3)
    init_key = jax.random.split(init_key, batch_size)
    # TODO: These all initialize in the same state, probably do not want this
    h0 = eqx.filter_vmap(model.initialize_carry)(init_key)

    model_key = jax.random.split(model_key, batch_size)

    _, y_preds = eqx.filter_vmap(model)(h0, (x, starts), model_key)
    # batch, time, feature
    y_pred = y_preds[:, -1]

    loss = mse(y_pred, y)
    l1 = l1_error(y_pred, y)
    return loss, {"loss": loss, "l1_error": l1}

def loss_classify_terminal_output(
    model: Module,
    x: Shaped[Array, "Batch Time Feature"],
    y: Shaped[Array, "Batch Classes"],
    key = None,
    decay = 0.01,
) -> Tuple[Shaped[Array, "1"], Dict[str, Array]]:
    """Given a sequence of inputs x1, ..., xn and predicted outputs y1p, ..., y1n,
    return the cross entropy loss between the true yn and predicted y1n.

    Args:
        model: memax.groups.Module
        x: (batch, time, in_feature)
        y: (batch, num_classes)

    Returns:
        loss: scalar
        info: dict
    """
    batch_size = x.shape[0]
    seq_len = x.shape[1]
    assert (
        x.shape[0] == y.shape[0]
    ), f"batch size mismatch: {x.shape[0]} != {y.shape[0]}"
    assert x.ndim == 3, f"expected 3d input, got {x.ndim}d"
    assert y.ndim == 2, f"expected 2d input, got {y.ndim}d"

    starts = jnp.zeros((batch_size, seq_len), dtype=bool)
    key, init_key, model_key = jax.random.split(key, 3)
    init_key = jax.random.split(init_key, batch_size)
    # TODO: These all initialize in the same state, probably do not want this
    h0 = eqx.filter_vmap(model.initialize_carry)(init_key)

    model_key = jax.random.split(model_key, batch_size)

    h, y_preds = eqx.filter_vmap(model)(h0, (x, starts), model_key)
    # batch, time, feature
    y_pred = y_preds[:, -1]
    loss = cross_entropy(y_pred, y)
    acc = accuracy(y_pred, y)
    return loss, {"loss": loss, "accuracy": acc}

def update_model(
    model: Module,
    loss_fn: Callable,
    opt: optax.GradientTransformation,
    opt_state: optax.OptState,
    x: Shaped[Array, "Batch ..."],
    y: Shaped[Array, "Batch ..."],
    key=None,
) -> Tuple[Module, optax.OptState, Dict[str, Array]]:
    """Update the model using the given loss function and optimizer."""
    grads, loss_info = eqx.filter_grad(loss_fn, has_aux=True)(model, x, y, key)
    updates, opt_state = opt.update(
        grads, opt_state, params=eqx.filter(model, eqx.is_inexact_array)
    )
    model = eqx.apply_updates(model, updates)
    return model, opt_state, loss_info

@eqx.filter_jit
def scan_one_epoch(
    model: Module,
    opt: optax.GradientTransformation,
    opt_state: optax.OptState,
    loss_fn: Callable,
    xs: Shaped[Array, "Datapoint ..."],
    ys: Shaped[Array, "Datapoint ..."],
    batch_size: int,
    batch_index: Shaped[Array, "Batch ..."],
    *,
    key: jax.random.PRNGKey,
) -> Tuple[Module, optax.OptState, Dict[str, Array]]:
    """Train a single epoch using the scan operator. Functions as a dataloader and train loop."""
    assert (
        xs.shape[0] == ys.shape[0]
    ), f"batch size mismatch: {xs.shape[0]} != {ys.shape[0]}"
    params, static = eqx.partition(model, eqx.is_array)

    def get_batch(x, y, step):
        """Returns a specific batch of size `batch_size` from `x` and `y`."""
        start = step * batch_size
        x_batch = jax.lax.dynamic_slice_in_dim(x, start, batch_size, 0)
        y_batch = jax.lax.dynamic_slice_in_dim(y, start, batch_size, 0)
        return x_batch, y_batch

    def inner(carry, index):
        params, opt_state, key = carry
        x, y = get_batch(xs, ys, index)
        key = jax.random.split(key)[0]
        model = eqx.combine(params, static)
        # JIT this otherwise it takes ages to compile the epoch
        params, opt_state, metrics = update_model(
            model, loss_fn, opt, opt_state, x, y, key=key
        )
        params, _ = eqx.partition(params, eqx.is_array)
        return (params, opt_state, key), metrics

    (params, opt_state, key), epoch_metrics = jax.lax.scan(
        inner,
        (params, opt_state, key),
        batch_index,
    )
    model = eqx.combine(params, static)
    return model, opt_state, epoch_metrics

def get_semigroups(
    recurrent_size: int,
    key: jax.random.PRNGKey,
) -> Dict[str, Module]:
    """Returns a dictionary containing all implemented semigroups.
    
    This returns the operator used in the scan, not the full recurrent cell. 
    """
    return {
        "PSpherical": PSphericalSemigroup(recurrent_size),
        "FFM": FFMSemigroup(recurrent_size, recurrent_size, recurrent_size, key=key),
        "FART": FARTSemigroup(recurrent_size),
        "LinearRNN": LinearRNNSemigroup(recurrent_size),
        "LRU": LRUSemigroup(recurrent_size),
        "S6": S6Semigroup(recurrent_size),
        "NMax": NMaxSemigroup(recurrent_size),
        "FWP": FWPSemigroup(recurrent_size),
        "DeltaNet": DeltaNetSemigroup(recurrent_size),
        "DeltaProduct": DeltaProductSemigroup(recurrent_size),
        "GDN": GDNSemigroup(recurrent_size),
        "Stack": StackSemigroup(recurrent_size, stack_size=4),
        "Attention": AttentionSemigroup(recurrent_size, window_size=4)
    }

def get_residual_memory_models(
    input: int,
    hidden: int,
    output: int,
    num_layers: int = 2,
    models: str = "all",
    *,
    key: jax.random.PRNGKey,
    layer_kwargs: Optional[Dict[str, Any]] = None,
    model_kwargs: Optional[Dict] = None,
) -> Dict[str, Module]:
    """Returns a dictionary of models, correponding to all semigroups and set actions.
    
    This returns a dictionary of models, each consisting of multiple recurrent cells 
    with residual and DenseNet connections between them.
    """
    layer_kwargs = layer_kwargs or {}
    model_kwargs = model_kwargs or {}
    layers = {
        # for debug
        "MLP": lambda recurrent_size, key: MLP(
            recurrent_size=recurrent_size, key=key, **layer_kwargs.get("MLP", {})
        ),
        # semigroups
        "NMax": lambda recurrent_size, key: NMax(
            recurrent_size=recurrent_size, key=key
        ),
        "FART": lambda recurrent_size, key: FART(
           hidden_size=recurrent_size, recurrent_size=round(recurrent_size ** 0.5), key=key, **layer_kwargs.get("FART", {})
        ),
        "FWP": lambda recurrent_size, key: FWP(
           hidden_size=recurrent_size, recurrent_size=round(recurrent_size ** 0.5), key=key, **layer_kwargs.get("FWP", {})
        ),
        "DeltaNet": lambda recurrent_size, key: DeltaNet(
           hidden_size=recurrent_size, recurrent_size=round(recurrent_size ** 0.5), key=key, **layer_kwargs.get("DeltaNet", {})
        ),
        "DeltaProduct": lambda recurrent_size, key: DeltaProduct(
           hidden_size=recurrent_size, recurrent_size=round(recurrent_size ** 0.5), rank=4, key=key, **layer_kwargs.get("DeltaProduct", {})
        ),
        "GDN": lambda recurrent_size, key: GDN(
           hidden_size=recurrent_size, recurrent_size=round(recurrent_size ** 0.5), key=key, **layer_kwargs.get("GDN", {})
        ),
        "FFM": lambda recurrent_size, key: FFM(
           hidden_size=recurrent_size, context_size=recurrent_size//4, trace_size=4, key=key, **layer_kwargs.get("FFM", {})
        ),
        "S6": lambda recurrent_size, key: S6(
            hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key, **layer_kwargs.get("S6", {})
        ),
        "PSpherical": lambda recurrent_size, key: PSpherical(
            recurrent_size=round(recurrent_size ** 0.5),
            hidden_size=recurrent_size,
            key=key,
            **layer_kwargs.get("PSpherical", {})
        ),
        "LRU": lambda recurrent_size, key: LRU(
            hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key, **layer_kwargs.get("LRU", {})
        ),
        "LinearRNN": lambda recurrent_size, key: LinearRecurrent(
            recurrent_size=recurrent_size, key=key, **layer_kwargs.get("LinearRNN", {})
        ),
        "Stack": lambda recurrent_size, key: Stack(
            recurrent_size=recurrent_size, stack_size=4, key=key, **layer_kwargs.get("Stack", {})
        ),
        "Attention": lambda recurrent_size, key: Attention(
            recurrent_size=recurrent_size, window_size=20, positional_embedding=None, key=key, **layer_kwargs.get("Attention", {})
        ),
        "Attention-RoPE": lambda recurrent_size, key: Attention(
            recurrent_size=recurrent_size, window_size=20, positional_embedding="rope", key=key, **layer_kwargs.get("Attention-RoPE", {})
        ),
        "Attention-ALiBi": lambda recurrent_size, key: Attention(
            recurrent_size=recurrent_size, window_size=20, positional_embedding="alibi", key=key, **layer_kwargs.get("Attention-ALiBi", {})
        ),
        # set actions
        "GRU": lambda recurrent_size, key: GRU(recurrent_size=recurrent_size, key=key, **layer_kwargs.get("GRU", {})),
        "Elman": lambda recurrent_size, key: Elman(
           hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key, **layer_kwargs.get("Elman", {})
        ),
        "ElmanReLU": lambda recurrent_size, key: Elman(
           hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key, activation=jax.nn.relu, **layer_kwargs.get("ElmanReLU", {})
        ),
        "Spherical": lambda recurrent_size, key: Spherical(
            hidden_size=recurrent_size, recurrent_size=recurrent_size, key=key, **layer_kwargs.get("Spherical", {})
        ),
        "MGU": lambda recurrent_size, key: MGU(recurrent_size=recurrent_size, key=key, **layer_kwargs.get("MGU", {})),
        "LSTM": lambda recurrent_size, key: LSTM(recurrent_size=recurrent_size, key=key, **layer_kwargs.get("LSTM", {})),
    }
    if models == "all":
        return {
            name: ResidualModel(
                make_layer_fn=fn,
                input_size=input,
                recurrent_size=hidden,
                output_size=output,
                num_layers=num_layers,
                key=key,
                **model_kwargs,
            )
            for name, fn in layers.items()
        }
    else:
        return {
            name: ResidualModel(
                make_layer_fn=layers[name],
                input_size=input,
                recurrent_size=hidden,
                output_size=output,
                num_layers=num_layers,
                key=key,
                **model_kwargs,
            )
            for name in models
        }


from beartype.typing import Optional, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.mtypes import Input, StartFlag
from memax.linen.groups import BinaryAlgebra, Semigroup, Resettable
from memax.linen.gras import GRAS
from memax.linen.scans import semigroup_scan

FARTRecurrentState = Tuple[Float[Array, "Key Value"], Float[Array, "Key"]]
FARTRecurrentStateWithReset = Tuple[FARTRecurrentState, StartFlag]


def phi(x):
    return 1 + jax.nn.elu(x)


class FARTSemigroup(Semigroup):
    """The Fast AutoRegressive Transformer semigroup (recurrent update) from https://arxiv.org/abs/2006.16236"""

    recurrent_size: int

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FARTRecurrentState:
        return (
            jnp.zeros((self.recurrent_size, self.recurrent_size)),
            jnp.zeros((self.recurrent_size)),
        )

    @jaxtyped(typechecker=typechecker)
    def zero_carry(self) -> FARTRecurrentState:
        return (
            jnp.zeros((self.recurrent_size, self.recurrent_size)),
            jnp.zeros((self.recurrent_size)),
        )

    @jaxtyped(typechecker=typechecker)
    @nn.compact
    def __call__(
        self,
        carry: FARTRecurrentState,
        input: FARTRecurrentState,
    ) -> FARTRecurrentState:
        (
            kv_sum,
            k_sum,
        ) = carry
        kv, k = input
        kv_sum = kv_sum + kv
        k_sum = k_sum + k
        return kv_sum, k


class FART(GRAS):
    """The Fast AutoRegressive Transformer from https://arxiv.org/abs/2006.16236.

    You might want to use this as a building block for a more complex model.
    """

    hidden_size: int
    recurrent_size: int
    algebra: BinaryAlgebra

    def setup(self):
        self.K = nn.Dense(self.recurrent_size)
        self.Q = nn.Dense(self.recurrent_size)
        self.V = nn.Dense(self.recurrent_size)
        self.output = nn.Dense(self.hidden_size)

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FARTRecurrentStateWithReset:
        emb, start = x
        k = phi(self.K(emb))
        v: Float[Array, "Time Feat"] = self.V(emb)
        kv = jnp.outer(k, v)
        return (kv, k), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: FARTRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        emb, start = x
        (kv_sum, k_sum), reset_flag = h
        q = phi(self.Q(emb))
        out = q @ kv_sum / (1e-6 + jnp.dot(k_sum, q))
        return self.output(out) + emb

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> FARTRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)

    @nn.nowrap
    def zero_carry(self) -> FARTRecurrentState:
        return self.algebra.zero_carry()

    @staticmethod
    def default_algebra(**kwargs):
        return Resettable(FARTSemigroup(**kwargs))

    @staticmethod
    def default_scan():
        return semigroup_scan



"""This module contains semigroup-based recurrent layers.
They are generally much faster than standard RNNs.

Each RNN type gets its own file.
+ `memax.equinox.semigroups.attn` provides dot-product attention layer. 
+ `memax.equinox.semigroups.delta` provides the DeltaNet layer.
+ `memax.equinox.semigroups.deltap` provides the DeltaProduct layer.
+ `memax.equinox.semigroups.stack` provides framestacking (sliding-window) as an RNN.
+ `memax.equinox.semigroups.fart` provides the Fast AutoRegressive Transformer layer.
+ `memax.equinox.semigroups.ffm` provides the Fast and Forgetful Memory layer.
+ `memax.equinox.semigroups.fwp` provides the Fast Weight Programmer layer.
+ `memax.equinox.semigroups.lru` provides the Linear Recurrent Unit layer.
+ `memax.equinox.semigroups.gdn` provides the Gated DeltaNet layer.
+ `memax.equinox.semigroups.lrnn` provides a basic linear recurrence. 
+ `memax.equinox.semigroups.mlp` provides an MLP (no memory) for completeness.
+ `memax.equinox.semigroups.s6` provides the Selective State Space Model (Mamba) layer.
+ `memax.equinox.semigroups.spherical` provides Rotational RNN layer (spherical projection).
"""



"""This module implements scans for executing recurrent models in Equinox.

They wrap JAX's scan and associative_scan functions, creating a unified interface
for executing recurrent models defined as binary algebras (set actions and semigroups).
It also provides some useful error checking to ensure the scans are used correctly.
"""
from beartype.typing import Callable

import flax.linen as nn
import jax
import jax.numpy as jnp

from memax.mtypes import RecurrentState
from memax.utils import debug_shape


def set_action_scan(
    set_action_op: Callable[[RecurrentState, RecurrentState], RecurrentState],
    state: RecurrentState,
    input: RecurrentState,
):
    """Update the recurrent state using a sequential scan.

    This works with virtually any recurrent model. You can even use it with semigroups,
    it will just be less efficient than an associative scan.
    """
    assert jax.tree.structure(state) == jax.tree.structure(
        input
    ), f"Mismatched structures passed to scan, {jax.tree.structure(state)} and {jax.tree.structure(input)}"

    def wrapped_set_action_op(set_action, carry, xs):
        out = set_action(carry, xs)
        return out, out  # Ensure the output carry matches the input structure

    _, new_state = nn.scan(
        target=wrapped_set_action_op,
        in_axes=0,
        variable_broadcast="params",
        split_rngs={"params": False},
    )(set_action_op, state, input)
    assert jax.tree.structure(new_state) == jax.tree.structure(
        input
    ), f"Mismatched structures returned from scan, {jax.tree.structure(input)} and {jax.tree.structure(new_state)}"
    return new_state


def semigroup_scan(
    semigroup_op: Callable[[RecurrentState, RecurrentState], RecurrentState],
    state: RecurrentState,
    input: RecurrentState,
    axis: int = 0,
) -> RecurrentState:
    r"""Update the recurrent state using an associative scan.

    The semigroup_op MUST be associative, i.e.,

    $f(a, f(b,c)) = f(f(a,b), c)$

    See https://en.wikipedia.org/wiki/Semigroup for information.
    """
    assert jax.tree.structure(state) == jax.tree.structure(
        input
    ), f"Mismatched structures passed to scan, {jax.tree.structure(state)} and {jax.tree.structure(input)}"

    # Both scan initial state and inputs must have the same numbers of dims
    # Usually, we have h0 = (features,) and x = (time, features)
    # Expand h0 = (1, features)
    singleton_state = jax.tree.map(lambda s: jnp.expand_dims(s, axis=axis), state)
    # Concatenate the previous state to the inputs and scan over the result
    # This ensures the previous recurrent state contributes to the current batch
    scan_inputs = jax.tree.map(
        lambda s, x: jnp.concatenate([s, x], axis=axis), singleton_state, input
    )

    new_state = jax.lax.associative_scan(
        fn=jax.vmap(semigroup_op), elems=scan_inputs, axis=axis
    )

    # The zeroth index corresponds to the previous recurrent state
    # We just use it to ensure continuity
    # We do not actually want to use these values, so slice them away
    new_state = jax.tree.map(
        lambda x: jax.lax.slice_in_dim(x, start_index=1, limit_index=None, axis=axis),
        new_state,
    )

    assert jax.tree.structure(new_state) == jax.tree.structure(
        input
    ), f"Mismatched structures returned from scan, {jax.tree.structure(input)} and {jax.tree.structure(new_state)}"
    assert all(
        jax.tree.leaves(jax.tree.map(lambda x, y: x.shape == y.shape, input, new_state))
    ), f"Shapes do not match {debug_shape(input)} and {debug_shape(new_state)}"

    return new_state



# https://github.com/NicolasZucchet/minimal-S6/blob/main/S6/model.py
from beartype.typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from equinox import nn
from beartype import beartype as typechecker
from jaxtyping import Array, Complex, Float, PRNGKeyArray, Scalar, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, Semigroup, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import semigroup_scan

S6RecurrentState = Tuple[Float[Array, "Recurrent"], Float[Array, "Recurrent"]]
S6RecurrentStateWithReset = Tuple[S6RecurrentState, StartFlag]


class S6Semigroup(Semigroup):
    """The S6 semigroup (recurrent update) from https://arxiv.org/abs/2312.00752.
    
    This is an S5/LRU recurrent update with a learnable timestep parameter. """

    recurrent_size: int

    def __init__(
        self,
        recurrent_size: int,
    ):
        self.recurrent_size = recurrent_size

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> S6RecurrentState:
        # Represent a diagonal matrix as a vector
        return (
            jnp.ones((self.recurrent_size,)),
            jnp.zeros((self.recurrent_size,))
        )

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: S6RecurrentState, input: S6RecurrentState
    ) -> S6RecurrentState:
        A_i, bu_i = carry
        A_j, bu_j = input
        return A_j * A_i, A_j * bu_i + bu_j


class S6(GRAS):
    """
    The S6 SSM with diagonal recurrence. We base this on the LRU, and add a trainable dt.

    You might want to use this as a building block for a more complex model.
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [S6RecurrentStateWithReset, S6RecurrentStateWithReset],
                S6RecurrentStateWithReset,
            ],
            S6RecurrentStateWithReset,
            S6RecurrentStateWithReset,
        ],
        S6RecurrentStateWithReset,
    ]
    A_log: Float[Array, "Recurrent"]
    B: nn.Linear
    C: nn.Linear
    dt: nn.Linear

    hidden_size: int  # input and output dimensions
    recurrent_size: int  # hidden state dimension

    def __init__(self, recurrent_size, hidden_size, key):
        keys = jax.random.split(key, 4)
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        unwrapped = S6Semigroup(recurrent_size)
        self.algebra = Resettable(unwrapped)
        self.scan = semigroup_scan

        self.A_log = jax.random.normal(keys[0], (self.recurrent_size,))
        self.B = nn.Linear(self.hidden_size, self.recurrent_size, key=keys[1])
        self.C = nn.Linear(self.recurrent_size, self.hidden_size, key=keys[2])
        self.dt = nn.Sequential([
            nn.Linear(self.hidden_size, self.recurrent_size, key=keys[3]),
            nn.Lambda(jax.nn.softplus)
        ])

    @jaxtyped(typechecker=typechecker)
    def forward_map(self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None):
        emb, start = x
        dt = self.dt(emb)
        A = -jnp.exp(self.A_log)
        A_bar = jnp.exp(dt * A)
        B = self.B(emb)
        # NOTE: A and B are diagonal so we can compute B_bar more simply than the mamba paper
        # Thankfully, inv(A) is just 1 / A if A is diagonal
        # Furthermore the dt's cancel: 1 / (dt A) with dt B
        B_bar = 1 / A * (A_bar - 1.0) * B
        Bu = B_bar * emb
        return (A_bar, Bu), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: S6RecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        state, reset_flag = h
        emb, start = x
        lambdas, lambda_x_Bu = state
        C = self.C(emb)
        out = C * lambda_x_Bu 
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> S6RecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from setuptools import find_packages, setup

setup(
    name="memax",
    version="0.1.0",
    author="Steven Morad",
    author_email="stevenmorad@gmail.com",
    description="Deep memory and sequence modeling in JAX",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/smorad/memax",
    packages=find_packages(),
    install_requires=[
        "jax",
        "jaxlib",
        "jaxtyping",
        "optax",
        "beartype",
    ],
    extras_require={
        'equinox': ['equinox'],
        # TODO: Update if flax fixes their shit
        'flax': [
            'flax',
            'please-downgrade-to-python-3.13-for-flax; python_version >= "3.14"', 
        ],
        'train': [
            'datasets',
            'tqdm',
            'pillow',
            'wandb',
        ],
        'all': [
            'equinox',
            'flax',
            'please-downgrade-to-python-3.13-for-flax; python_version >= "3.14"',
            # train
            'datasets',
            'tqdm',
            'pillow',
            'wandb',
        ],
        'test': [
            'pytest',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



from beartype.typing import Optional, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.mtypes import Input, StartFlag
from memax.linen.groups import BinaryAlgebra, SetAction, Resettable
from memax.linen.gras import GRAS
from memax.linen.scans import set_action_scan

GRURecurrentState = Float[Array, "Recurrent"]
GRURecurrentStateWithReset = Tuple[GRURecurrentState, StartFlag]


class GRUSetAction(SetAction):
    """
    The Gated Recurrent Unit set action

    Paper: https://arxiv.org/abs/1406.1078
    """

    recurrent_size: int

    def setup(self):
        self.U_z = nn.Dense(
            self.recurrent_size,
            use_bias=False,
        )
        self.U_r = nn.Dense(
            self.recurrent_size,
            use_bias=False,
        )
        self.U_h = nn.Dense(
            self.recurrent_size,
            use_bias=False,
        )

        self.W_z = nn.Dense(self.recurrent_size)
        self.W_r = nn.Dense(self.recurrent_size)
        self.W_h = nn.Dense(self.recurrent_size)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: GRURecurrentState, input: Float[Array, "Recurrent"]
    ) -> GRURecurrentState:
        z = jax.nn.sigmoid(self.W_z(input) + self.U_z(carry))
        r = jax.nn.sigmoid(self.W_r(input) + self.U_r(carry))
        h_hat = jax.nn.tanh(self.W_h(input) + self.U_h(r * carry))
        out = (1 - z) * carry + z * h_hat
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentState:
        return jnp.zeros((self.recurrent_size,))

    @nn.nowrap
    def zero_carry(self) -> GRURecurrentState:
        return jnp.zeros((self.recurrent_size,))


class GRU(GRAS):
    """
    The Gated Recurrent Unit

    Paper: https://arxiv.org/abs/1406.1078
    """

    recurrent_size: int

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentStateWithReset:
        emb, start = x
        return emb, start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: GRURecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        z, reset_flag = h
        emb, start = x
        return z

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> GRURecurrentStateWithReset:
        return self.algebra.initialize_carry(key)

    @nn.nowrap
    def zero_carry(self) -> GRURecurrentStateWithReset:
        return self.algebra.zero_carry()

    @staticmethod
    def default_algebra(**kwargs):
        return Resettable(GRUSetAction(**kwargs))

    @staticmethod
    def default_scan():
        return set_action_scan



from beartype.typing import Callable, List, Optional, Tuple

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from equinox import nn
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.equinox.groups import BinaryAlgebra, SetAction, Module, Resettable
from memax.equinox.gras import GRAS
from memax.mtypes import Input, StartFlag
from memax.equinox.scans import set_action_scan

SphericalRecurrentState = Float[Array, "Recurrent"]
SphericalRecurrentStateWithReset = Tuple[SphericalRecurrentState, StartFlag]


class SphericalMagma(SetAction):
    """
    The RotRNN (recurrent update) from https://arxiv.org/abs/2407.07239

    However, this is implemented in a less efficient manner (sequential)
    """

    recurrent_size: int
    project: nn.Linear
    initial_state: jax.Array

    def __init__(self, recurrent_size: int, sequence_length: int = 1024, *, key):
        self.recurrent_size = recurrent_size
        proj_size = int(self.recurrent_size * (self.recurrent_size - 1) / 2)
        self.project = nn.Linear(recurrent_size, proj_size, key=key)
        self.initial_state = jnp.ones((self.recurrent_size,))

    @jaxtyped(typechecker=typechecker)
    def rot(self, z: Array) -> Array:
        q = self.project(z)
        A = jnp.zeros((self.recurrent_size, self.recurrent_size))
        tri_idx = jnp.triu_indices_from(A, 1)
        A = A.at[tri_idx].set(q)
        A = A - A.T
        R = jax.scipy.linalg.expm(A)
        return R

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, carry: SphericalRecurrentState, input: SphericalRecurrentState
    ) -> SphericalRecurrentState:
        R = self.rot(input)
        return R @ carry

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentState:
        return self.initial_state / jnp.linalg.norm(self.initial_state)


class Spherical(GRAS):
    """The Spherical RNN from https://arxiv.org/abs/2407.07239
    
    However, this is implemented in a less efficient manner (sequential)
    than the spherical semigroup.
    """

    algebra: BinaryAlgebra
    scan: Callable[
        [
            Callable[
                [SphericalRecurrentStateWithReset, SphericalRecurrentStateWithReset],
                SphericalRecurrentStateWithReset,
            ],
            SphericalRecurrentStateWithReset,
            SphericalRecurrentStateWithReset,
        ],
        SphericalRecurrentStateWithReset,
    ]
    W_y: nn.Linear
    recurrent_size: int
    hidden_size: int

    def __init__(self, recurrent_size, hidden_size, key):
        self.recurrent_size = recurrent_size
        self.hidden_size = hidden_size
        keys = jax.random.split(key)
        self.algebra = Resettable(SphericalMagma(recurrent_size, key=keys[0]))
        self.scan = set_action_scan
        self.W_y = nn.Linear(recurrent_size, hidden_size, key=keys[1])

    @jaxtyped(typechecker=typechecker)
    def forward_map(
        self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentStateWithReset:
        emb, start = x
        return emb, start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: SphericalRecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.hidden_size}"]:
        z, reset_flag = h
        emb, start = x
        return self.W_y(z)

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> SphericalRecurrentStateWithReset:
        return self.algebra.initialize_carry(key)



from functools import partial
from beartype.typing import Optional, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray, Shaped

from memax.mtypes import Input, RecurrentState, ResetRecurrentState, StartFlag
from memax.utils import debug_shape


class Module(nn.Module):
    r"""
    The base module for memory/sequence models.

    A module $m$ maps a recurrent state and inputs to an output recurrent state and outputs.
    We always include a binary start flag in the inputs.

    $ m: H \times X^n \times \\{0, 1\\}^n \mapsto H^n \times Y^n $


    The start flag signifies the beginning of a new sequence. For example,

        [1, 2, 3, 4, 5]
        [0, 0, 1, 0, 1]

    Denotes three distinct input sequences: [1, 2], [3, 4], and [5]. With `Resettable`, these sequences
    will not interfere with each other. As such, you can concatenate multiple sequences
    into a single sequence for efficient processing.
    """

    @nn.compact
    def __call__(self, h: RecurrentState, x: Input) -> RecurrentState:
        r"""Applies $\bullet$ to the inputs to return an updated recurrent state."""
        raise NotImplementedError

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> RecurrentState:
        r"""Returns the initial recurrent state $h_0$."""
        raise NotImplementedError

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        r"""Returns zeros in the shape of the initial recurrent state $h_0$."""
        raise NotImplementedError


class BinaryAlgebra(Module):
    r"""An binary algebraic structure (e.g., semigroup, set action) that maps two inputs to an output.

    You must define an initial state $h_0$ and a binary operator $\bullet$.
    """

    @nn.compact
    def __call__(self, carry: RecurrentState, input: RecurrentState) -> RecurrentState:
        raise NotImplementedError

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> RecurrentState:
        raise NotImplementedError

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        raise NotImplementedError


class SetAction(BinaryAlgebra):
    r"""
    A set action, a form of binary algebra that we execute using scans.

    $ \bullet: H \times Z \mapsto H $.
    """

    @nn.compact
    def __call__(self, carry: RecurrentState, input: RecurrentState) -> RecurrentState:
        raise NotImplementedError

    def initialize_carry(self, batch_shape: Tuple[int, ...] = ()) -> RecurrentState:
        raise NotImplementedError

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        raise NotImplementedError


class Semigroup(BinaryAlgebra):
    r"""A semigroup, as defined in https://en.wikipedia.org/wiki/Semigroup.

    A semigroup is a constrained form of set action, where:
    1. The action and recurrent state spaces are identical, i.e., $Z = H$.
    2. The binary operator $\bullet$ is associative.

    $ \bullet: H \times H \mapsto H $

    $ (a \bullet b) \bullet c = a \bullet (b \bullet c) $.
    """

    @nn.compact
    def __call__(self, carry: RecurrentState, input: RecurrentState) -> RecurrentState:
        raise NotImplementedError

    def initialize_carry(self, batch_shape: Tuple[int, ...] = ()) -> RecurrentState:
        raise NotImplementedError

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        raise NotImplementedError


class Resettable(BinaryAlgebra):
    r"""A wrapper that resets the recurrent state upon beginning a new sequence.

    This is a binary algebra defined on another binary algebra.

    For set actions this is

    $ \circ: (H \times \\{0, 1\\}) \times (Z \times \\{0, 1\\}) \mapsto (H \times \\{0, 1\\}) $

    while for semigroups this is

    $ \circ: (H \times \\{0, 1\\}) \times (H \times \\{0, 1\\}) \mapsto (H \times \\{0, 1\\}) $.
    """

    algebra: BinaryAlgebra

    @nn.compact
    def __call__(self, carry: ResetRecurrentState, input: ResetRecurrentState):
        assert jax.tree.structure(carry) == jax.tree.structure(
            input
        ), f"Mismatched structures passed to algebra, {jax.tree.structure(carry)} and {jax.tree.structure(input)}"
        states, prev_carry_reset_flag = carry
        xs, start = input

        def reset_state(
            start: StartFlag,
            current_state: RecurrentState,
            initial_state: RecurrentState,
        ):
            # Expand to reset all dims of state: [1, B, 1, ...]
            assert initial_state.ndim == current_state.ndim
            out = current_state * jnp.logical_not(start) + initial_state * start
            return out

        initial_states = self.algebra.initialize_carry()
        states = jax.tree.map(partial(reset_state, start), states, initial_states)
        out = self.algebra(states, xs)
        carry_reset_flag = jnp.logical_or(start, prev_carry_reset_flag)
        to_return = out, carry_reset_flag
        assert jax.tree.structure(carry) == jax.tree.structure(
            to_return
        ), f"Mismatched structures passed from algebra,\n{jax.tree.structure(carry)} and\n{jax.tree.structure(out)}"
        assert all(
            jax.tree.leaves(jax.tree.map(lambda x, y: x.shape == y.shape, states, out))
        ), f"Shapes do not match\n{debug_shape(states)} and\n{debug_shape(out)}"

        return to_return

    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> RecurrentState:
        return self.algebra.initialize_carry(key), jnp.zeros((), dtype=bool)

    @nn.nowrap
    def zero_carry(self) -> RecurrentState:
        return self.algebra.zero_carry(), jnp.zeros((), dtype=bool)

    def __getattr__(self, name):
        # Delegate attribute access to the algebra instance if not found in Resettable
        if hasattr(self.algebra, name):
            return getattr(self.algebra, name)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )



"""Test all models on a simple 'remember the first input in the sequence' task"""
import pytest
import jax
import jax.numpy as jnp
import optax
from functools import partial

from memax.linen.train_utils import get_residual_memory_models


def get_desired_accuracies():
    return {
        "LRU": 0.999,
        "S6": 0.999,
        "FART": 0.999,
        "GRU": 0.999,
    }

def ce_loss(y_hat, y):
    return -jnp.mean(jnp.sum(y * jax.nn.log_softmax(y_hat, axis=-1), axis=-1))

@pytest.mark.parametrize("model_name, model", get_residual_memory_models(
        8, 4 - 1, 
    ).items())
def test_initial_input(
    model_name, model, epochs=4000, num_seqs=5, seq_len=20, input_dims=4
):
    timesteps = num_seqs * seq_len
    seq_idx = jnp.array([seq_len * i for i in range(num_seqs)])
    start = jnp.zeros((timesteps,), dtype=bool).at[seq_idx].set(True)
    opt = optax.adam(learning_rate=3e-3)

    # init model
    key = jax.random.PRNGKey(0)
    dummy_x = jax.random.randint(key, (timesteps,), 0, input_dims - 1)
    dummy_x = jax.nn.one_hot(dummy_x, input_dims - 1)
    dummy_x = jnp.concatenate([dummy_x, start.astype(jnp.float32).reshape(-1, 1)], axis=-1)
    dummy_h = model.zero_carry()
    dummy_starts = jnp.zeros(dummy_x.shape[0], dtype=bool)
    params = model.init(key, dummy_h, (dummy_x, dummy_starts))
    init_carry_fn = partial(model.apply, method="initialize_carry")
    apply_fn = model.apply
    state = opt.init(params)

    def error(params, key):
        h = init_carry_fn(params) 
        x = jax.random.randint(key, (timesteps,), 0, input_dims - 1)
        x = jax.nn.one_hot(x, input_dims - 1)
        x = jnp.concatenate([x, start.astype(jnp.float32).reshape(-1, 1)], axis=-1)
        y = jnp.repeat(x[seq_idx, :-1], seq_len, axis=0)

        _, y_hat = apply_fn(params, h, (x, start))
        loss = ce_loss(y_hat, y)
        accuracy = jnp.mean(jnp.argmax(y, axis=-1) == jnp.argmax(y_hat, axis=-1))
        return loss, {"loss": loss, "accuracy": accuracy}

    loss_fn = jax.jit(jax.grad(error, has_aux=True))
    losses = []
    accuracies = []
    for epoch in range(epochs):
        key, _ = jax.random.split(key)
        grads, loss_info = loss_fn(params, key)
        updates, state = jax.jit(opt.update)(grads, state)
        params = optax.apply_updates(params, updates)
        losses.append(loss_info["loss"])
        accuracies.append(loss_info["accuracy"])

    losses, accuracies = jnp.stack(losses), jnp.stack(accuracies)
    losses = losses[-100:].mean()
    accuracies = accuracies[-100:].mean()
    print(f"{model_name} mean accuracy: {accuracies:0.3f}")
    assert (
        accuracies >= get_desired_accuracies()[model_name]
    ), f"Failed {model_name}, expected {get_desired_accuracies()[model_name]}, got {accuracies}"

    # Verify recurrent mode works well too
    def rerror(params, key):
        h = init_carry_fn(params) 
        x = jax.random.randint(key, (timesteps,), 0, input_dims - 1)
        x = jax.nn.one_hot(x, input_dims - 1)
        x = jnp.concatenate([x, start.astype(jnp.float32).reshape(-1, 1)], axis=-1)
        y = jnp.repeat(x[seq_idx, :-1], seq_len, axis=0)
        y_hats = []

        for t in range(timesteps):
            h, y_hat = apply_fn(params, h, (x[t : t + 1], start[t : t + 1]))
            h = model.latest_recurrent_state(h)
            y_hats.append(y_hat)

        y_hat = jnp.concatenate(y_hats, axis=0)
        loss = ce_loss(y_hat, y)
        accuracy = jnp.mean(jnp.argmax(y, axis=-1) == jnp.argmax(y_hat, axis=-1))
        return loss, {"loss": loss, "accuracy": accuracy}

    _, r_metrics = rerror(params, key)
    assert (
        r_metrics['accuracy']>= get_desired_accuracies()[model_name]
    ), f"Failed {model_name} (recurrent mode), expected {get_desired_accuracies()[model_name]}, got {r_metrics['accuracy']}"


if __name__ == "__main__":
    test_initial_input()



# https://github.com/NicolasZucchet/minimal-S6/blob/main/lru/model.py
from functools import partial
from beartype.typing import Optional, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Float, PRNGKeyArray, Shaped, jaxtyped

from memax.mtypes import Input, StartFlag
from memax.linen.groups import Semigroup, Resettable
from memax.linen.gras import GRAS
from memax.linen.scans import semigroup_scan

S6RecurrentState = Tuple[Float[Array, "Recurrent"], Float[Array, "Recurrent"]]
S6RecurrentStateWithReset = Tuple[S6RecurrentState, StartFlag]


class S6Semigroup(Semigroup):
    """The diagonal S6 semigroup (recurrent update) from https://arxiv.org/abs/2312.00752.
    
    This is a diagonal S5/LRU recurrent update with a learnable timestep parameter. """

    recurrent_size: int

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> S6RecurrentState:
        # Represent a diagonal matrix as a vector
        return (
            jnp.ones((self.recurrent_size,)),
            jnp.zeros((self.recurrent_size)),
        )

    @nn.nowrap
    def zero_carry(self) -> S6RecurrentState:
        return (
            jnp.zeros((self.recurrent_size,)),
            jnp.zeros((self.recurrent_size)),
        )

    @jaxtyped(typechecker=typechecker)
    @nn.compact
    def __call__(
        self, carry: S6RecurrentState, input: S6RecurrentState
    ) -> S6RecurrentState:
        # Ax + Bu, but A is diagonal, and we treat it as a vector
        # So we can be more efficient by writing Ax as vec(A) * x
        A_i, bu_i = carry
        A_j, bu_j = input
        return A_j * A_i, A_j * bu_i + bu_j


class S6(GRAS):
    """
    The diagonal S6 SSM, an SSM with a trainable dt.

    You might want to use this as a building block for a more complex model.
    """

    hidden_size: int  # output dimensions
    recurrent_size: int  # hidden state dimension

    def setup(self):
        self.A_log = self.param(
            "A_log",
            jax.random.normal,
            (self.recurrent_size,),
        )
        self.B = nn.Dense(self.recurrent_size)
        self.C = nn.Dense(self.hidden_size)
        self.dt = nn.Sequential([
            nn.Dense(self.recurrent_size),
            jax.nn.softplus
        ])

    @jaxtyped(typechecker=typechecker)
    def forward_map(self, x: Input, key: Optional[Shaped[PRNGKeyArray, ""]] = None):
        emb, start = x
        dt = self.dt(emb)
        A = -jnp.exp(self.A_log)
        A_bar = jnp.exp(dt * A)
        B = self.B(emb)
        # NOTE: A and B are diagonal so we can compute B_bar more simply than the mamba paper
        # Thankfully, inv(A) is just 1 / A if A is diagonal
        # Furthermore the dt's cancel: 1 / (dt A) with dt B
        B_bar = 1 / A * (A_bar - 1.0) * B
        Bu = B_bar * emb
        return (A_bar, Bu), start

    @jaxtyped(typechecker=typechecker)
    def backward_map(
        self,
        h: S6RecurrentStateWithReset,
        x: Input,
        key: Optional[Shaped[PRNGKeyArray, ""]] = None,
    ) -> Float[Array, "{self.recurrent_size}"]:
        state, reset_flag = h
        emb, start = x
        lambdas, lambda_x_Bu = state
        C = self.C(emb)
        out = C * lambda_x_Bu 
        return out

    @jaxtyped(typechecker=typechecker)
    def initialize_carry(
        self, key: Optional[Shaped[PRNGKeyArray, ""]] = None
    ) -> S6RecurrentStateWithReset:
        return self.algebra.initialize_carry(key)

    @nn.nowrap
    def zero_carry(self) -> S6RecurrentStateWithReset:
        return self.algebra.zero_carry()

    @staticmethod
    def default_algebra(**kwargs):
        return Resettable(S6Semigroup(**kwargs))

    @staticmethod
    def default_scan():
        return semigroup_scan
